#!/usr/bin/env python
"""Flags shared across multiple commands."""

from typing import List, Optional

from absl import flags


def define_null_marker(
    flag_values: flags.FlagValues,
) -> flags.FlagHolder[Optional[str]]:
  return flags.DEFINE_string(
      'null_marker',
      None,
      'An optional custom string that will represent a NULL value'
      'in CSV External table data.',
      flag_values=flag_values,
  )


def define_null_markers(
    flag_values: flags.FlagValues,
) -> flags.FlagHolder[Optional[List[str]]]:
  return flags.DEFINE_list(
      'null_markers',
      [],
      'An optional list of custom strings that will represent a NULL value'
      'in CSV External table data.',
      flag_values=flag_values,
  )


def define_time_zone(
    flag_values: flags.FlagValues,
) -> flags.FlagHolder[Optional[str]]:
  return flags.DEFINE_string(
      'time_zone',
      None,
      'Default time zone that will apply when parsing timestamp values that'
      ' have no specific time zone. For example, "America/Los_Angeles".',
      flag_values=flag_values,
  )


def define_date_format(
    flag_values: flags.FlagValues,
) -> flags.FlagHolder[Optional[str]]:
  return flags.DEFINE_string(
      'date_format',
      None,
      'Format elements that define how the DATE values are formatted in the'
      ' input files. For example, "MM/DD/YYYY".',
      flag_values=flag_values,
  )


def define_datetime_format(
    flag_values: flags.FlagValues,
) -> flags.FlagHolder[Optional[str]]:
  return flags.DEFINE_string(
      'datetime_format',
      None,
      'Format elements that define how the DATETIME values are formatted in'
      ' the input files. For example, "MM/DD/YYYY HH24:MI:SS.FF3".',
      flag_values=flag_values,
  )


def define_time_format(
    flag_values: flags.FlagValues,
) -> flags.FlagHolder[Optional[str]]:
  return flags.DEFINE_string(
      'time_format',
      None,
      'Format elements that define how the TIME values are formatted in the'
      ' input files. For example, "HH24:MI:SS.FF3".',
      flag_values=flag_values,
  )


def define_timestamp_format(
    flag_values: flags.FlagValues,
) -> flags.FlagHolder[Optional[str]]:
  return flags.DEFINE_string(
      'timestamp_format',
      None,
      'Format elements that define how the TIMESTAMP values are formatted in'
      ' the input files. For example, "MM/DD/YYYY HH24:MI:SS.FF3".',
      flag_values=flag_values,
  )


def define_source_column_match(
    flag_values: flags.FlagValues,
) -> flags.FlagHolder[Optional[str]]:
  return flags.DEFINE_enum(
      'source_column_match',
      None,
      ['POSITION', 'NAME'],
      'Controls the strategy used to match loaded columns to the schema in CSV'
      ' External table data. Options include:'
      '\n POSITION: matches by position. This option assumes that the columns'
      ' are ordered the same way as the schema.'
      '\n NAME: matches by name. This option reads the header row as column'
      ' names and reorders columns to match the field names in the schema.',
      flag_values=flag_values,
  )


def define_parquet_map_target_type(
    flag_values: flags.FlagValues,
) -> flags.FlagHolder[Optional[str]]:
  return flags.DEFINE_enum(
      'parquet_map_target_type',
      None,
      ['ARRAY_OF_STRUCT'],
      'Specifies the parquet map type. If it is equal to ARRAY_OF_STRUCT,'
      ' then a map_field will be represented with a repeated struct (that has'
      ' key and value fields).',
      flag_values=flag_values,
  )


def define_timestamp_target_precision(
    flag_values: flags.FlagValues,
) -> flags.FlagHolder[Optional[List[str]]]:
  return flags.DEFINE_list(
      'timestamp_target_precision',
      [],
      'Precisions (maximum number of total digits in base 10) for second of'
      ' TIMESTAMP types that are allowed to the destination table for'
      ' autodetection mode.'
      '\n'
      '\n Available for the formats: CSV.'
      '\n'
      '\n For the CSV Format, Possible values include:'
      '\n Not Specified, [], or [6]: timestamp(6) for all auto detected'
      ' TIMESTAMP columns.'
      '\n [6, 12]: timestamp(6) for all auto detected TIMESTAMP columns that'
      ' have less than 6 digits of subseconds. timestamp(12) for all auto'
      ' detected TIMESTAMP columns that have more than 6 digits of subseconds.'
      '\n [12]: timestamp(12) for all auto detected TIMESTAMP columns.'
      '\n'
      '\n The order of the elements in this array is ignored.'
      '\n Inputs that have higher precision than the highest target precision'
      ' will be truncated.',
      flag_values=flag_values,
  )


def define_reservation_id_for_a_job(
    flag_values: flags.FlagValues,
) -> flags.FlagHolder[Optional[str]]:
  return flags.DEFINE_string(
      'reservation_id',
      None,
      'Reservation ID used when executing the job. Reservation should be in the'
      'format of project_id:reservation_id, project_id:location.reservation_id,'
      'or reservation_id',
      flag_values=flag_values,
  )


def define_event_driven_schedule(
    flag_values: flags.FlagValues,
) -> flags.FlagHolder[Optional[str]]:
  return flags.DEFINE_string(
      'event_driven_schedule',
      None,
      'Event driven schedule in json format. Example:'
      ' --event_driven_schedule=\'{"pubsub_subscription":'
      ' "projects/project-id/subscriptions/subscription-id"}\'. This flag'
      ' should not be used with --schedule, --no_auto_scheduling,'
      ' --schedule_start_time or --schedule_end_time.',
      flag_values=flag_values,
  )


def define_use_full_timestamp(
    flag_values: flags.FlagValues,
) -> flags.FlagHolder[bool]:
  return flags.DEFINE_boolean(
      'use_full_timestamp',
      False,
      'Use full precision ISO8601 string representation for timestamp values'
      ' in the result.',
      flag_values=flag_values,
  )


def define_parent_group(
    flag_values: flags.FlagValues,
) -> flags.FlagHolder[Optional[str]]:
  return flags.DEFINE_string(
      'parent_group',
      None,
      'The resource name of the parent reservation group. It can be a full '
      'resource name or just the reservation group ID.',
      flag_values=flag_values,
  )



# -*- coding: utf-8 -*- #
# Copyright 2026 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Translation rule for networking features."""

from collections.abc import Mapping, Sequence
import json
from typing import Any


def translate_network_features(
    input_data: Mapping[str, Any],
) -> Sequence[str]:
  """Translate networking features.

  Args:
    input_data: Flattened dictionary of the parsed app.yaml file.

  Returns:
    A list of strings representing the flags for Cloud Run.
  """
  output_flags = []
  if 'network.name' in input_data:
    output_flags.append(f'--network={input_data["network.name"]}')
  if 'network.subnetwork_name' in input_data:
    output_flags.append(f'--subnet={input_data["network.subnetwork_name"]}')
  if 'network.instance_tag' in input_data:
    output_flags.append(f'--network-tags={input_data["network.instance_tag"]}')
  if input_data.get('network.session_affinity'):
    output_flags.append('--session-affinity')

  forwarded_ports = input_data.get('network.forwarded_ports') or input_data.get(
      'network.forwardedPorts'
  )
  if forwarded_ports:
    output_flags.append(f'--port={forwarded_ports[0]}')

  instance_ip_mode = input_data.get(
      'network.instance_ip_mode'
  ) or input_data.get('network.instanceIpMode')
  if instance_ip_mode:
    if instance_ip_mode == 'external':
      output_flags.append('--ingress=all')
    elif instance_ip_mode == 'internal':
      output_flags.append('--ingress=internal')

  return output_flags


def update_service_yaml_with_network(
    service_yaml: dict[str, Any],
    input_data: Mapping[str, Any],
) -> None:
  """Update the service_yaml dict with networking settings.

  Args:
    service_yaml: The dictionary representation of the Cloud Run service YAML to
      be updated.
    input_data: Flattened dictionary of the parsed app.yaml file, containing
      networking configurations.
  """

  # 1. Service level annotations
  service_metadata = service_yaml.setdefault('metadata', {})
  service_annotations = service_metadata.setdefault('annotations', {})

  instance_ip_mode = input_data.get(
      'network.instance_ip_mode'
  ) or input_data.get('network.instanceIpMode')
  if instance_ip_mode == 'external':
    service_annotations['run.googleapis.com/ingress'] = 'all'
  elif instance_ip_mode == 'internal':
    service_annotations['run.googleapis.com/ingress'] = 'internal'

  # 2. Revision level metadata & annotations
  spec = service_yaml.setdefault('spec', {})
  template = spec.setdefault('template', {})
  template_metadata = template.setdefault('metadata', {})
  template_annotations = template_metadata.setdefault('annotations', {})

  if input_data.get('network.session_affinity'):
    template_annotations['run.googleapis.com/sessionAffinity'] = 'true'

  # 3. Direct VPC (Network, Subnet, and Tags)
  network_interface = {}
  if network_name := input_data.get('network.name'):
    network_interface['network'] = network_name
  if subnetwork_name := input_data.get('network.subnetwork_name'):
    network_interface['subnetwork'] = subnetwork_name
  if tag_data := input_data.get('network.instance_tag'):
    network_interface['tags'] = (
        tag_data if isinstance(tag_data, list) else [tag_data]
    )

  if network_interface:
    template_annotations['run.googleapis.com/network-interfaces'] = json.dumps(
        [network_interface]
    )

  vpc_connector = input_data.get('vpc_access_connector.name') or input_data.get(
      'vpcAccessConnector.name'
  )
  if vpc_connector:
    template_annotations['run.googleapis.com/vpc-access-connector'] = (
        vpc_connector
    )

  vpc_egress = input_data.get(
      'vpc_access_connector.egress_setting'
  ) or input_data.get('vpcAccessConnector.egressSetting')
  if vpc_egress:
    template_annotations['run.googleapis.com/vpc-access-egress'] = vpc_egress

  # 4. Container level settings
  template_spec = template.setdefault('spec', {})
  containers = template_spec.setdefault('containers', [{}])
  container = containers[0]

  forwarded_ports = input_data.get('network.forwarded_ports') or input_data.get(
      'network.forwardedPorts'
  )
  if forwarded_ports:
    container_ports = container.setdefault('ports', [])
    container_ports.append({'containerPort': int(forwarded_ports[0])})



# -*- coding: utf-8 -*- #
# Copyright 2026 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""List command to list security profile groups."""

from googlecloudsdk.api_lib.network_security.security_profile_groups import spg_api
from googlecloudsdk.calliope import base
from googlecloudsdk.command_lib.network_security import spg_flags

_DETAILED_HELP = {
    'DESCRIPTION': """

          List all Security Profile Groups in the specified location.

        """,
    'EXAMPLES': """
          To list Security Profile Groups in specifed location, run:

              $ {command} --location=global

        """,
}

_FORMAT = """\
table(
    name.basename():label=NAME
)
"""

_PROJECT_SCOPE_SUPPORTED_TRACKS = (
    base.ReleaseTrack.ALPHA,
    base.ReleaseTrack.BETA,
    base.ReleaseTrack.GA,
)


@base.DefaultUniverseOnly
@base.ReleaseTracks(
    base.ReleaseTrack.ALPHA, base.ReleaseTrack.BETA, base.ReleaseTrack.GA
)
class ListProfileGroup(base.ListCommand):
  """List Security Profile Groups."""

  @classmethod
  def Args(cls, parser):
    project_scope_supported = (
        cls.ReleaseTrack() in _PROJECT_SCOPE_SUPPORTED_TRACKS
    )
    parser.display_info.AddFormat(_FORMAT)
    parser.display_info.AddUriFunc(spg_flags.MakeGetUriFunc(cls.ReleaseTrack()))
    spg_flags.AddLocationResourceArg(
        parser=parser,
        help_text='Parent resource for the list operation.',
        required=True,
        project_scope_supported=project_scope_supported,
    )

  def Run(self, args):
    result = args.CONCEPTS.location.Parse()
    parent_ref = result.result

    project_scoped = (
        result.concept_type.name
        == spg_flags.PROJECT_LOCATION_RESOURCE_COLLECTION
    )
    client = spg_api.Client(self.ReleaseTrack(), project_scoped)

    return client.ListSecurityProfileGroups(
        parent_ref.RelativeName(), page_size=args.page_size
    )


ListProfileGroup.detailed_help = _DETAILED_HELP



# -*- coding: utf-8 -*- #
# Copyright 2021 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Flags and helpers for the Datastream related commands."""


_MYSQL_EXCLUDED_OBJECTS_HELP_TEXT = """\
  Path to a YAML (or JSON) file containing the MySQL data sources to avoid backfilling.

  The JSON file is formatted as follows, with camelCase field naming:

  ```
    {
      "mysqlDatabases": [
        {
          "database":"sample_database",
          "mysqlTables": [
            {
              "table": "sample_table",
              "mysqlColumns": [
                {
                  "column": "sample_column",
                }
                ]
            }
          ]
        }
      ]
    }
  ```
"""

_ORACLE_EXCLUDED_OBJECTS_HELP_TEXT = """\
  Path to a YAML (or JSON) file containing the Oracle data sources to avoid backfilling.

  The JSON file is formatted as follows, with camelCase field naming:

  ```
    {
      "oracleSchemas": [
        {
          "schema": "SAMPLE",
          "oracleTables": [
            {
              "table": "SAMPLE_TABLE",
              "oracleColumns": [
                {
                  "column": "COL",
                }
              ]
            }
          ]
        }
      ]
    }
  ```
"""

_POSTGRESQL_EXCLUDED_OBJECTS_HELP_TEXT = """\
  Path to a YAML (or JSON) file containing the PostgreSQL data sources to avoid backfilling.

  The JSON file is formatted as follows, with camelCase field naming:

  ```
    {
      "postgresqlSchemas": [
        {
          "schema": "SAMPLE",
          "postgresqlTables": [
            {
              "table": "SAMPLE_TABLE",
              "postgresqlColumns": [
                {
                  "column": "COL",
                }
              ]
            }
          ]
        }
      ]
    }
  ```
"""

_SQLSERVER_EXCLUDED_OBJECTS_HELP_TEXT = """\
  Path to a YAML (or JSON) file containing the SQL Server data sources to avoid backfilling.

  The JSON file is formatted as follows, with camelCase field naming:

  ```
    {
      "schemas": [
        {
          "schema": "SAMPLE",
          "tables": [
            {
              "table": "SAMPLE_TABLE",
              "columns": [
                {
                  "column": "COL",
                }
              ]
            }
          ]
        }
      ]
    }
  ```
"""

_SALESFORCE_EXCLUDED_OBJECTS_HELP_TEXT = """\
  Path to a YAML (or JSON) file containing the Salesforce data sources to avoid backfilling.

  The JSON file is formatted as follows, with camelCase field naming:

  ```
    {
      "objects": [
        {
          "objectName": "SAMPLE",
        },
        {
          "objectName": "SAMPLE2",
        }
      ]
    }
  ```
"""

_SPANNER_EXCLUDED_OBJECTS_HELP_TEXT = """\
  Path to a YAML (or JSON) file containing the Spanner data sources to avoid backfilling.

  The JSON file is formatted as follows, with camelCase field naming:

  ```
    {
      "schemas": [
        {
          "schema": "SAMPLE_SCHEMA",
          "tables": [
            {
              "table": "SAMPLE_TABLE",
              "columns": [
                {
                  "column": "SAMPLE_COLUMN",
                }
              ]
            }
          ]
        }
      ]
    }
  ```
"""

_MONGODB_EXCLUDED_OBJECTS_HELP_TEXT = """\
  Path to a YAML (or JSON) file containing the MongoDB data sources to avoid backfilling.

  The JSON file is formatted as follows, with camelCase field naming:

  ```
    {
      "databases": [
        {
          "database":"sample_database",
          "collections": [
            {
              "collection": "sample_collection",
              "fields": [
                {
                  "field": "sample_field",
                }
              ]
            }
          ]
        }
      ]
    }
  ```
"""


_SAAS_EXCLUDED_OBJECTS_HELP_TEXT = """\
  Path to a YAML (or JSON) file containing the SaaS data sources to avoid backfilling.

  The JSON file is formatted as follows, with camelCase field naming:

  ```
    {
      "objects": [
        {
          "objectName": "Account",
        }
      ]
    }
  ```
"""


def AddRuleSetsFlag(parser):
  """Adds a --rule-sets flag to the given parser."""
  help_text = """Path to a JSON file containing a list of rule sets
  to be applied to the stream.

  The JSON file is formatted as follows, with camelCase field naming:

  ```
  [
    {
      "objectFilter": {
        "sourceObjectIdentifier": {
          "oracleIdentifier": {
            "schema": "schema1",
            "table": "table1"
          }
        }
      },
      "customizationRules": [
        {
          "bigqueryClustering": {
            "columns": ["COL_A"]
          }
        }
      ]
    },
    {
      "objectFilter": {
        "sourceObjectIdentifier": {
          "oracleIdentifier": {
            "schema": "schema2",
            "table": "table2"
          }
        }
      },
      "customizationRules": [
        {
          "bigqueryPartitioning": {
            "timeUnitPartition": {
              "column": "TIME_COL",
              "partitioningTimeGranularity": "PARTITIONING_TIME_GRANULARITY_DAY"
            }
          }
        }
      ]
    }
  ]
  ```
"""
  parser.add_argument('--rule-sets', help=help_text)


def AddDisplayNameFlag(parser, required=True):
  """Adds a --display-name flag to the given parser."""
  help_text = """Friendly name for the stream."""
  parser.add_argument('--display-name', help=help_text, required=required)


def AddUpdateMaskFlag(parser):
  """Adds a --update-mask flag to the given parser."""
  help_text = """Used to specify the fields to be overwritten in the stream resource by the update.
  If the update mask is used, then a field will be overwritten only if it is in the mask. If the user does not provide a mask then all fields will be overwritten.
  This is a comma-separated list of fully qualified names of fields, written as snake_case or camelCase. Example: "display_name, source_config.oracle_source_config"."""
  parser.add_argument('--update-mask', help=help_text)


def AddStateFlag(parser):
  """Adds a --state flag to the given parser."""
  help_text = """Stream state, can be set to: "RUNNING" or "PAUSED"."""
  parser.add_argument('--state', help=help_text)


def AddValidationGroup(parser, verb):
  """Adds a --validate-only or --force flag to the given parser."""
  validation_group = parser.add_group(mutex=True)
  validation_group.add_argument(
      '--validate-only',
      help="""Only validate the stream, but do not %s any resources.
      The default is false.""" % verb.lower(),
      action='store_true',
      default=False)
  validation_group.add_argument(
      '--force',
      help="""%s the stream without validating it.""" % verb,
      action='store_true',
      default=False)


def AddBackfillStrategyGroup(parser, required=True):
  """Adds a --backfiill-all or --backfill-none flag to the given parser."""
  backfill_group = parser.add_group(required=required, mutex=True)
  backfill_group.add_argument(
      '--backfill-none',
      help="""Do not automatically backfill any objects. This flag is equivalent
      to selecting the Manual backfill type in the Google Cloud console.""",
      action='store_true')
  backfill_all_group = backfill_group.add_group()
  backfill_all_group.add_argument(
      '--backfill-all',
      help="""Automatically backfill objects included in the stream source
      configuration. Specific objects can be excluded. This flag is equivalent
      to selecting the Automatic backfill type in the Google Cloud console.""",
      action='store_true')
  backfill_all_excluded_objects = backfill_all_group.add_group(mutex=True)
  backfill_all_excluded_objects.add_argument(
      '--oracle-excluded-objects', help=_ORACLE_EXCLUDED_OBJECTS_HELP_TEXT
  )
  backfill_all_excluded_objects.add_argument(
      '--mysql-excluded-objects', help=_MYSQL_EXCLUDED_OBJECTS_HELP_TEXT
  )
  backfill_all_excluded_objects.add_argument(
      '--postgresql-excluded-objects',
      help=_POSTGRESQL_EXCLUDED_OBJECTS_HELP_TEXT,
  )
  backfill_all_excluded_objects.add_argument(
      '--sqlserver-excluded-objects',
      help=_SQLSERVER_EXCLUDED_OBJECTS_HELP_TEXT,
  )
  backfill_all_excluded_objects.add_argument(
      '--salesforce-excluded-objects',
      help=_SALESFORCE_EXCLUDED_OBJECTS_HELP_TEXT,
  )
  backfill_all_excluded_objects.add_argument(
      '--spanner-excluded-objects',
      help=_SPANNER_EXCLUDED_OBJECTS_HELP_TEXT,
  )
  backfill_all_excluded_objects.add_argument(
      '--mongodb-excluded-objects',
      help=_MONGODB_EXCLUDED_OBJECTS_HELP_TEXT,
  )
  backfill_all_excluded_objects.add_argument(
      '--saas-excluded-objects',
      help=_SAAS_EXCLUDED_OBJECTS_HELP_TEXT,
  )



# -*- coding: utf-8 -*- #
# Copyright 2020 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Create a certificate."""


from googlecloudsdk.api_lib.cloudkms import cryptokeyversions
from googlecloudsdk.api_lib.privateca import base as privateca_base
from googlecloudsdk.api_lib.privateca import certificate_utils
from googlecloudsdk.api_lib.privateca import request_utils
from googlecloudsdk.calliope import base
from googlecloudsdk.calliope import exceptions
from googlecloudsdk.calliope.concepts import deps
from googlecloudsdk.command_lib.privateca import flags
from googlecloudsdk.command_lib.privateca import key_generation
from googlecloudsdk.command_lib.privateca import pem_utils
from googlecloudsdk.command_lib.privateca import resource_args
from googlecloudsdk.command_lib.util.args import labels_util
from googlecloudsdk.command_lib.util.concepts import concept_parsers
from googlecloudsdk.command_lib.util.concepts import presentation_specs
from googlecloudsdk.core import log
from googlecloudsdk.core.util import files

_KEY_OUTPUT_HELP = """The path where the generated private key file should be written (in PEM format).

Note: possession of this key file could allow anybody to act as this certificate's
subject. Please make sure that you store this key file in a secure location at all
times, and ensure that only authorized users have access to it."""


def _ReadCsr(csr_file):
  try:
    return files.ReadFileContents(csr_file)
  except (files.Error, OSError, IOError):
    raise exceptions.BadFileException(
        "Could not read provided CSR file '{}'.".format(csr_file)
    )


def _WritePemChain(pem_cert, issuing_chain, cert_file):
  try:
    pem_chain = [pem_cert] + issuing_chain
    files.WriteFileContents(cert_file, pem_utils.PemChainForOutput(pem_chain))
  except (files.Error, OSError, IOError):
    raise exceptions.BadFileException(
        "Could not write certificate to '{}'.".format(cert_file)
    )


@base.ReleaseTracks(base.ReleaseTrack.GA)
@base.UniverseCompatible
class Create(base.CreateCommand):
  r"""Create a new certificate.

  ## EXAMPLES

  To create a certificate using a CSR:

      $ {command} frontend-server-tls \
        --issuer-pool=my-pool --issuer-location=us-west1 \
        --csr=./csr.pem \
        --cert-output-file=./cert.pem \
        --validity=P30D

  To create a certificate using a client-generated key:

      $ {command} frontend-server-tls \
        --issuer-pool=my-pool --issuer-location=us-west1 \
        --generate-key \
        --key-output-file=./key \
        --cert-output-file=./cert.pem \
        --dns-san=www.example.com \
        --use-preset-profile=leaf_server_tls
  """

  @staticmethod
  def Args(parser):
    persistence_group = parser.add_group(
        mutex=True, required=True, help='Certificate persistence options.'
    )
    base.Argument(
        '--cert-output-file',
        help=(
            'The path where the resulting PEM-encoded certificate chain file'
            ' should be written (ordered from leaf to root).'
        ),
        required=False,
    ).AddToParser(persistence_group)
    base.Argument(
        '--validate-only',
        help=(
            'If this flag is set, the certificate resource will not be'
            ' persisted and the returned certificate will not contain the'
            ' pem_certificate field.'
        ),
        action='store_true',
        default=False,
        required=False,
    ).AddToParser(persistence_group)

    flags.AddValidityFlag(parser, 'certificate', 'P30D', '30 days')
    flags.AddRequestedNotBeforeTimeFlag(parser)
    labels_util.AddCreateLabelsFlags(parser)

    cert_generation_group = parser.add_group(
        mutex=True, required=True, help='Certificate generation method.'
    )

    csr_group = cert_generation_group.add_group(
        help='To issue a certificate from a CSR use the following:',
    )
    base.Argument(
        '--csr', help='A PEM-encoded certificate signing request file path.',
        required=True,
    ).AddToParser(csr_group)

    base.Argument(
        '--rdn-sequence-subject',
        help=(
            'If this value is set then the issued certificate will use the '
            'subject found in the CSR preserving the exact RDN sequence.'
        ),
        action='store_true',
    ).AddToParser(csr_group)

    non_csr_group = cert_generation_group.add_group(
        help='Alternatively, you may describe the certificate and key to use.'
    )
    key_group = non_csr_group.add_group(
        mutex=True,
        required=True,
        help=(
            'To describe the key that will be used for this certificate, use '
            'one of the following options.'
        ),
    )
    key_generation_group = key_group.add_group(
        help='To generate a new key pair, use the following:'
    )
    base.Argument(
        '--generate-key',
        help=(
            'Use this flag to have a new RSA-2048 private key securely'
            ' generated on your machine.'
        ),
        action='store_const',
        const=True,
        default=False,
        required=True,
    ).AddToParser(key_generation_group)
    base.Argument(
        '--key-output-file', help=_KEY_OUTPUT_HELP, required=True
    ).AddToParser(key_generation_group)
    base.Argument(
        '--ca',
        help=(
            'The name of an existing certificate authority to use for issuing'
            ' the certificate. If omitted, a certificate authority will be will'
            ' be chosen from the CA pool by the service on your behalf.'
        ),
        required=False,
    ).AddToParser(parser)
    subject_group = non_csr_group.add_group(
        help='The subject names for the certificate.', required=True
    )
    flags.AddSubjectFlags(subject_group)
    x509_parameters_group = non_csr_group.add_group(
        mutex=True, help='The x509 configuration used for this certificate.'
    )
    flags.AddInlineX509ParametersFlags(
        x509_parameters_group, is_ca_command=False, default_max_chain_length=0
    )
    flags.AddUsePresetProfilesFlag(x509_parameters_group)
    flags.AddSubjectKeyIdFlag(parser)

    cert_arg = 'CERTIFICATE'
    concept_parsers.ConceptParser(
        [
            presentation_specs.ResourcePresentationSpec(
                cert_arg,
                resource_args.CreateCertResourceSpec(
                    cert_arg, [Create._GenerateCertificateIdFallthrough()]
                ),
                'The name of the certificate to issue. If the certificate ID '
                'is omitted, a random identifier will be generated according '
                'to the following format: {YYYYMMDD}-{3 random alphanumeric '
                'characters}-{3 random alphanumeric characters}. The '
                'certificate ID is not required when the issuing CA pool is in '
                'the DevOps tier.',
                required=True,
            ),
            presentation_specs.ResourcePresentationSpec(
                '--template',
                resource_args.CreateCertificateTemplateResourceSpec(
                    'certificate_template'
                ),
                'The name of a certificate template to use for issuing this '
                'certificate, if desired. A template may overwrite parts of '
                'the certificate request, and the use of certificate templates '
                "may be required and/or regulated by the issuing CA Pool's CA "
                'Manager. The specified template must be in the same location '
                'as the issuing CA Pool.',
                required=False,
                prefixes=True,
            ),
            presentation_specs.ResourcePresentationSpec(
                '--kms-key-version',
                resource_args.CreateKmsKeyVersionResourceSpec(),
                'An existing KMS key version backing this certificate.',
                group=key_group,
            ),
        ],
        command_level_fallthroughs={
            '--template.location': ['CERTIFICATE.issuer-location']
        },
    ).AddToParser(parser)

    # The only time a resource is returned is when args.validate_only is set.
    parser.display_info.AddFormat('yaml(certificateDescription)')

  @classmethod
  def _GenerateCertificateIdFallthrough(cls):
    cls.id_fallthrough_was_used = False

    def FallthroughFn():
      cls.id_fallthrough_was_used = True
      return certificate_utils.GenerateCertId()

    return deps.Fallthrough(
        function=FallthroughFn,
        hint='certificate id will default to an automatically generated id',
        active=False,
        plural=False,
    )

  def _ValidateArgs(self, args):
    """Validates the command-line args."""
    if args.IsSpecified('use_preset_profile') and args.IsSpecified('template'):
      raise exceptions.OneOfArgumentsRequiredException(
          ['--use-preset-profile', '--template'],
          (
              'To create a certificate, please specify either a preset profile '
              'or a certificate template.'
          ),
      )

    resource_args.ValidateResourceIsCompleteIfSpecified(args, 'kms_key_version')

  @classmethod
  def _PrintWarningsForUnpersistedCert(cls, args):
    """Prints warnings if certain command-line args are used for an unpersisted cert."""
    unused_args = []
    if not cls.id_fallthrough_was_used:
      unused_args.append('certificate ID')
    if args.IsSpecified('labels'):
      unused_args.append('labels')

    if unused_args:
      names = ', '.join(unused_args)
      verb = 'was' if len(unused_args) == 1 else 'were'
      log.warning(
          '{names} {verb} specified but will not be used since the '
          'issuing CA pool is in the DevOps tier, which does not expose '
          'certificate lifecycle.'.format(names=names, verb=verb)
      )

  def _GetPublicKey(self, args):
    """Fetches the public key associated with a non-CSR certificate request, as UTF-8 encoded bytes."""
    kms_key_version = args.CONCEPTS.kms_key_version.Parse()
    if args.generate_key:
      private_key, public_key = key_generation.RSAKeyGen(2048)
      key_generation.ExportPrivateKey(args.key_output_file, private_key)
      return public_key
    elif kms_key_version:
      public_key_response = cryptokeyversions.GetPublicKey(kms_key_version)
      # bytes(..) requires an explicit encoding in PY3.
      return bytes(public_key_response.pem, 'utf-8')
    else:
      # This should not happen because of the required arg group, but protects
      # in case of future additions.
      raise exceptions.OneOfArgumentsRequiredException(
          ['--csr', '--generate-key', '--kms-key-version'],
          (
              'To create a certificate, please specify either a CSR, the'
              ' --generate-key flag to create a new key, or the'
              ' --kms-key-version flag to use an existing KMS key.'
          ),
      )

  def _GenerateCertificateConfig(self, request, args):
    public_key = self._GetPublicKey(args)

    config = self.messages.CertificateConfig()
    config.publicKey = self.messages.PublicKey()
    config.publicKey.key = public_key
    config.publicKey.format = self.messages.PublicKey.FormatValueValuesEnum.PEM
    config.subjectConfig = flags.ParseSubjectFlags(args)
    config.x509Config = flags.ParseX509Parameters(args, is_ca_command=False)
    config.subjectKeyId = flags.ParseSubjectKeyId(args, self.messages)
    return config

  def Run(self, args):
    self.client = privateca_base.GetClientInstance(api_version='v1')
    self.messages = privateca_base.GetMessagesModule(api_version='v1')

    self._ValidateArgs(args)

    cert_ref = args.CONCEPTS.certificate.Parse()
    labels = labels_util.ParseCreateArgs(
        args, self.messages.Certificate.LabelsValue
    )

    request = (
        self.messages.PrivatecaProjectsLocationsCaPoolsCertificatesCreateRequest()
    )
    request.certificate = self.messages.Certificate()
    request.certificateId = cert_ref.Name()
    request.certificate.lifetime = flags.ParseValidityFlag(args)
    if args.IsSpecified('requested_not_before_time'):
      request.certificate.requestedNotBeforeTime = (
          args.requested_not_before_time
      )
    request.certificate.labels = labels
    request.parent = cert_ref.Parent().RelativeName()
    request.requestId = request_utils.GenerateRequestId()
    request.validateOnly = args.validate_only
    if args.IsSpecified('ca'):
      request.issuingCertificateAuthorityId = args.ca

    template_ref = args.CONCEPTS.template.Parse()
    if template_ref:
      if template_ref.locationsId != cert_ref.locationsId:
        raise exceptions.InvalidArgumentException(
            '--template',
            'The certificate template must be in the same location as the '
            'issuing CA Pool.',
        )
      request.certificate.certificateTemplate = template_ref.RelativeName()

    if args.csr:
      request.certificate.pemCsr = _ReadCsr(args.csr)
      if args.rdn_sequence_subject:
        request.certificate.subjectMode = (
            self.messages.Certificate.SubjectModeValueValuesEnum.RDN_SEQUENCE
        )
    else:
      request.certificate.config = self._GenerateCertificateConfig(
          request, args
      )

    certificate = self.client.projects_locations_caPools_certificates.Create(
        request
    )

    # Validate-only certs don't have a resource name or pem certificate.
    if args.validate_only:
      return certificate

    status_message = 'Created Certificate'

    if certificate.name:
      status_message += ' [{}]'.format(certificate.name)
    else:
      Create._PrintWarningsForUnpersistedCert(args)

    if certificate.pemCertificate:
      status_message += ' and saved it to [{}]'.format(args.cert_output_file)
      _WritePemChain(
          certificate.pemCertificate,
          certificate.pemCertificateChain,
          args.cert_output_file,
      )

    status_message += '.'
    log.status.Print(status_message)



# -*- coding: utf-8 -*- #
# Copyright 2023 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Flags for backup-dr commands."""

import argparse
import itertools
import textwrap
from typing import Any

import frozendict
from googlecloudsdk.calliope import actions
from googlecloudsdk.calliope import arg_parsers
from googlecloudsdk.calliope import base
from googlecloudsdk.calliope.concepts import concepts
from googlecloudsdk.calliope.concepts import deps
from googlecloudsdk.command_lib.backupdr import util
from googlecloudsdk.command_lib.util.concepts import concept_parsers
from googlecloudsdk.command_lib.util.concepts import presentation_specs
from googlecloudsdk.core import properties


MONTH_OPTIONS = frozendict.frozendict({
    'JAN': 'JANUARY',
    'FEB': 'FEBRUARY',
    'MAR': 'MARCH',
    'APR': 'APRIL',
    'MAY': 'MAY',
    'JUN': 'JUNE',
    'JUL': 'JULY',
    'AUG': 'AUGUST',
    'SEP': 'SEPTEMBER',
    'OCT': 'OCTOBER',
    'NOV': 'NOVEMBER',
    'DEC': 'DECEMBER',
})
DAY_OPTIONS = frozendict.frozendict({
    'MON': 'MONDAY',
    'TUE': 'TUESDAY',
    'WED': 'WEDNESDAY',
    'THU': 'THURSDAY',
    'FRI': 'FRIDAY',
    'SAT': 'SATURDAY',
    'SUN': 'SUNDAY',
})
RECURRENCE_OPTIONS = ('HOURLY', 'DAILY', 'WEEKLY', 'MONTHLY', 'YEARLY')
WEEK_OPTIONS = ('FIRST', 'SECOND', 'THIRD', 'FOURTH', 'LAST')
BACKUP_RULE_COMMON_HELP_TEXT = """
Parameters for the backup rule include:

*rule-id*::: Name of the backup rule. The name must be unique and
start with a lowercase letter followed by up to 62 lowercase letters,
numbers, or hyphens.

*retention-days*::: Duration for which backup data should be
retained. It must be defined in "days". The value should be greater
than or equal to the enforced retention period set for the backup vault.

*recurrence*::: Frequency for the backup schedule. It must be either:
HOURLY, DAILY, WEEKLY, MONTHLY or YEARLY.

*backup-window-start*::: Start time of the interval during which
backup jobs should be executed. It can be defined as backup-window-start=2,
that means backup window starts at 2 a.m. The start time and end time must
have an interval of 6 hours.

*backup-window-end*::: End time of the interval during which backup
jobs should be executed. It can be defined as backup-window-end=14,
that means backup window ends at 2 p.m. The start time and end time
must have an interval of 6 hours.

Jobs are queued at the beginning of the window and will be marked as
`SKIPPED` if they do not start by the end time. Jobs that are
in progress will not be canceled at the end time.

*time-zone*::: The time zone to be used for the backup schedule.
The value must exist in the
[IANA tz database](https://www.iana.org/time-zones).
The default value is UTC. E.g., Europe/Paris

::: Following flags are mutually exclusive:

*hourly-frequency*::: Frequency for hourly backups. An hourly
frequency of 2 means backup jobs will run every 2 hours from start
time till the end time defined. The hourly frequency must be between
4 and 23. The value is needed only if recurrence type is HOURLY.

*days-of-week*::: Days of the week when the backup job should be
executed. The value is needed if recurrence type is WEEKLY.
E.g., MONDAY,TUESDAY

*days-of-month*::: Days of the month when the backup job should
be executed. The value is needed only if recurrence type is YEARLY.
E.g.,"1,5,14"

*months*::: Month for the backup schedule. The value is needed only if
recurrence type is YEARLY. E.g., JANUARY, MARCH

*week-day-of-month*::: Recurring day of the week in the month or
year when the backup job should be executed. E.g. FIRST-SUNDAY, THIRD-MONDAY.
The value can only be provided if the recurrence type is MONTHLY or YEARLY.
Allowed values for the number of week - FIRST, SECOND, THIRD, FOURTH, LAST.
Allowed values for days of the week - MONDAY to SUNDAY.

::: E.g., "rule-id=sample-daily-rule,recurrence=WEEKLY,backup-window-start=2,backup-window-end=14,retention-days=20,days-of-week='SUNDAY MONDAY'"
"""


def BackupVaultAttributeConfig():
  return concepts.ResourceParameterAttributeConfig(
      name='backup-vault', help_text='The ID of the Backup Vault.'
  )


def DataSourceAttributeConfig():
  return concepts.ResourceParameterAttributeConfig(
      name='data-source', help_text='The ID of the Data Source.'
  )


def GetManagementServerResourceSpec():
  return concepts.ResourceSpec(
      'backupdr.projects.locations.managementServers',
      resource_name='Management Server',
      locationsId=LocationAttributeConfig(),
      projectsId=concepts.DEFAULT_PROJECT_ATTRIBUTE_CONFIG,
      disable_auto_completers=False,
  )


def BackupPlanAssociationProjectAttributeConfig():
  return concepts.ResourceParameterAttributeConfig(
      name='workload-project',
      help_text='Cloud project id for the {resource}.',
      fallthroughs=[
          deps.ArgFallthrough('--project'),
          deps.PropertyFallthrough(properties.VALUES.core.project),
      ],
  )


def GetBackupPlanResourceSpec():
  return concepts.ResourceSpec(
      'backupdr.projects.locations.backupPlans',
      resource_name='Backup Plan',
      locationsId=LocationAttributeConfig(),
      projectsId=concepts.DEFAULT_PROJECT_ATTRIBUTE_CONFIG,
      disable_auto_completers=False,
  )


def GetBackupPlanAssociationResourceSpec():
  return concepts.ResourceSpec(
      'backupdr.projects.locations.backupPlanAssociations',
      resource_name='Backup Plan Association',
      locationsId=LocationAttributeConfig(),
      projectsId=BackupPlanAssociationProjectAttributeConfig(),
      disable_auto_completers=False,
  )


def GetBackupResourceSpec():
  return concepts.ResourceSpec(
      'backupdr.projects.locations.backupVaults.dataSources.backups',
      resource_name='Backup',
      locationsId=LocationAttributeConfig(),
      projectsId=concepts.DEFAULT_PROJECT_ATTRIBUTE_CONFIG,
      backupVaultsId=BackupVaultAttributeConfig(),
      dataSourcesId=DataSourceAttributeConfig(),
      disable_auto_completers=False,
  )


def GetDataSourceResourceSpec():
  return concepts.ResourceSpec(
      'backupdr.projects.locations.backupVaults.dataSources',
      resource_name='Data Source',
      locationsId=LocationAttributeConfig(),
      projectsId=concepts.DEFAULT_PROJECT_ATTRIBUTE_CONFIG,
      backupVaultsId=BackupVaultAttributeConfig(),
      dataSourcesId=DataSourceAttributeConfig(),
      disable_auto_completers=False,
  )


def AddManagementServerResourceArg(parser, help_text):
  """Adds an argument for management server to parser."""
  name = 'management_server'
  concept_parsers.ConceptParser.ForResource(
      name,
      GetManagementServerResourceSpec(),
      help_text,
      required=True,
  ).AddToParser(parser)


def AddBackupPlanAssociationResourceArg(parser, help_text):
  """Adds an argument for backup plan association to parser."""
  name = 'backup_plan_association'
  concept_parsers.ConceptParser.ForResource(
      name,
      GetBackupPlanAssociationResourceSpec(),
      help_text,
      required=True,
  ).AddToParser(parser)


def AddBackupResourceArg(parser, help_text):
  """Adds an argument for backup to parser."""
  name = 'backup'
  concept_parsers.ConceptParser.ForResource(
      name,
      GetBackupResourceSpec(),
      help_text,
      required=True,
  ).AddToParser(parser)


def AddCreateBackupPlanAssociationFlags(parser):
  """Adds flags required to create a backup plan association."""
  concept_parsers.ConceptParser(
      [
          presentation_specs.ResourcePresentationSpec(
              'BACKUP_PLAN_ASSOCIATION',
              GetBackupPlanAssociationResourceSpec(),
              'Name of the backup plan association to be created. Once the'
              " backup plan association is created, this name can't be changed."
              ' The name must be unique for a project and location. To create'
              " backup plan associations in a project that's different from the"
              ' backup plan, use the --workload-project flag.',
              required=True,
          ),
          presentation_specs.ResourcePresentationSpec(
              '--backup-plan',
              GetBackupPlanResourceSpec(),
              'The backup plan to be applied to the resource. E.g.,'
              ' projects/sample-project/locations/us-central1/backupPlans/sample-backup-plan',
              # This hides the location flag for backup plan.
              flag_name_overrides={
                  'location': '',
              },
              required=True,
          ),
      ],
      command_level_fallthroughs={
          '--backup-plan.location': ['BACKUP_PLAN_ASSOCIATION.location'],
      },
  ).AddToParser(parser)

  parser.add_argument(
      '--resource',
      required=True,
      type=str,
      help=(
          'The resource to which the backup plan is to be applied. E.g.,'
          ' projects/sample-project/zones/us-central1-a/instances/sample-instance'
      ),
  )

  AddResourceType(
      parser,
      textwrap.dedent("""\
      Type of resource to which the backup plan should be applied.

      For example:
      * `compute.<UNIVERSE_DOMAIN>/Instance` for Compute Engine instances.
      * `file.<UNIVERSE_DOMAIN>/Instance` for Filestore instances."""),
  )


def AddUpdateBackupPlanAssociationFlags(parser):
  """Adds flags required to update a backup plan association."""
  concept_parsers.ConceptParser(
      [
          presentation_specs.ResourcePresentationSpec(
              'BACKUP_PLAN_ASSOCIATION',
              GetBackupPlanAssociationResourceSpec(),
              'Backup plan association to be updated. To update'
              " backup plan associations in a project that's different from the"
              ' backup plan, use the --workload-project flag.',
              required=True,
          ),
          presentation_specs.ResourcePresentationSpec(
              '--backup-plan',
              GetBackupPlanResourceSpec(),
              'Name of the specific backup plan to be applied to the backup'
              ' plan association. E.g.,'
              ' projects/sample-project/locations/us-central1/backupPlans/'
              'sample-backup-plan',
              # This hides the location flag for backup plan.
              flag_name_overrides={
                  'location': '',
              },
              required=True,
          ),
      ],
      command_level_fallthroughs={
          '--backup-plan.location': ['BACKUP_PLAN_ASSOCIATION.location'],
      },
  ).AddToParser(parser)


def AddTriggerBackupFlags(parser):
  """Adds flags required to create a backup plan association."""
  concept_parsers.ConceptParser(
      [
          presentation_specs.ResourcePresentationSpec(
              'BACKUP_PLAN_ASSOCIATION',
              GetBackupPlanAssociationResourceSpec(),
              'Name of an existing backup plan association to use for creating'
              ' an on-demand backup.',
              required=True,
          ),
      ],
  ).AddToParser(parser)

  group = parser.add_mutually_exclusive_group()
  group.add_argument(
      '--backup-rule-id',
      required=False,
      type=str,
      help=(
          'Name of an existing backup rule to use for creating an on-demand'
          ' backup.'
      ),
  )

  group.add_argument(
      '--custom-retention-days',
      required=False,
      type=int,
      help=(
          'Duration for which backup data will be retained.'
      ),
  )

  base.Argument(
      '--labels',
      metavar='KEY=VALUE',
      type=arg_parsers.ArgDict(value_type=str),
      action=arg_parsers.UpdateAction,
      help='Labels to be applied to the backup.').AddToParser(parser)


def AddNetwork(parser, required=False):
  """Adds a positional network argument to parser.

  Args:
    parser: argparse.Parser: Parser object for command line inputs.
    required: Whether or not --network is required.
  """
  parser.add_argument(
      '--network',
      required=required,
      type=str,
      help=(
          'Name of an existing VPC network with private service access'
          ' configured in the format -'
          ' projects/<project>/global/networks/<network>. This VPC network'
          ' allows the management console to communicate with all'
          ' backup/recovery appliances and requires a minimum IP range of /23.'
          ' This value cannot be changed after you deploy the management'
          " server. If you don't have private service access, configure one."
          ' [Learn more]'
          ' (https://cloud.google.com/vpc/docs/configure-private-services-access)'
      ),
      action=actions.DeprecationAction('--network', removed=False),
  )


def AddBackupVaultResourceArg(parser, help_text):
  """Adds an argument for backup vault to parser."""
  name = 'backup_vault'
  concept_parsers.ConceptParser.ForResource(
      name,
      GetBackupVaultResourceSpec(),
      help_text,
      required=True,
  ).AddToParser(parser)


def GetBackupVaultResourceSpec():
  return concepts.ResourceSpec(
      'backupdr.projects.locations.backupVaults',
      resource_name='Backup Vault',
      locationsId=LocationAttributeConfig(),
      projectsId=concepts.DEFAULT_PROJECT_ATTRIBUTE_CONFIG,
      disable_auto_completers=False,
  )


def LocationAttributeConfig(arg_name='location', default=None):
  """Creates location attribute config."""
  fallthroughs = []
  if default:
    fallthroughs.append(
        deps.Fallthrough(
            lambda: default,
            'Defaults to all locations',
        )
    )

  return concepts.ResourceParameterAttributeConfig(
      name=arg_name,
      fallthroughs=fallthroughs,
      help_text='The location of the {resource}.',
  )


def GetLocationResourceSpec(resource_name='location', default=None):
  return concepts.ResourceSpec(
      'backupdr.projects.locations',
      resource_name=resource_name,
      locationsId=LocationAttributeConfig(default=default),
      projectsId=concepts.DEFAULT_PROJECT_ATTRIBUTE_CONFIG,
  )


def AddLocationResourceArg(parser, help_text, default=None, required=True):
  """Adds an argument for location to parser."""
  name = '--location'
  override = None
  if default == 'global':
    override = {'location': ''}

  concept_parsers.ConceptParser.ForResource(
      name,
      GetLocationResourceSpec(default=default),
      help_text,
      flag_name_overrides=override,
      required=required,
  ).AddToParser(parser)


def AddNoAsyncFlag(parser):
  """Adds the --no-async flag to the given parser."""
  help_text = 'Wait for the operation in progress to complete.'
  parser.add_argument('--no-async', action='store_true', help=help_text)


def AddIgnoreInactiveDatasourcesFlag(parser):
  """Adds a --ignore-inactive-datasources flag to the given parser."""
  help_text = (
      'If set, the following restrictions against deletion of'
      ' the backup vault instance can be overridden:'
      ' * deletion of a backup vault instance containing no backups,'
      'but still contains empty datasources.'
  )
  parser.add_argument(
      '--ignore-inactive-datasources', action='store_true', help=help_text
  )


def AddIgnoreBackupPlanReferencesFlag(parser):
  """Adds a --ignore-backup-plan-references flag to the given parser."""
  help_text = (
      'If set, the following restrictions against deletion of'
      ' the backup vault instance can be overridden:'
      ' * deletion of a backup vault instance being actively referenced'
      ' by a backup plan.'
  )
  parser.add_argument(
      '--ignore-backup-plan-references', action='store_true', help=help_text
  )


def AddForceUpdateFlag(parser):
  """Adds a --force-update flag to the given parser."""
  help_text = (
      'If set, allow update to extend the minimum enforced retention for backup'
      ' vault. This overrides the restriction against conflicting retention'
      ' periods. This conflict may occur when the expiration schedule defined'
      ' by the associated backup plan is shorter than the minimum retention set'
      ' by the backup vault.'
  )
  parser.add_argument('--force-update', action='store_true', help=help_text)


def AddBackupPlanResourceArg(parser, help_text):
  """Adds an argument for backup plan to parser."""
  name = 'backup_plan'
  concept_parsers.ConceptParser.ForResource(
      name,
      GetBackupPlanResourceSpec(),
      help_text,
      required=True,
  ).AddToParser(parser)


def AddBackupPlanResourceArgWithBackupVault(parser, help_text):
  """Adds an argument for backup plan & backup vault to parser."""
  concept_parsers.ConceptParser(
      [
          presentation_specs.ResourcePresentationSpec(
              'BACKUP_PLAN',
              GetBackupPlanResourceSpec(),
              help_text,
              required=True,
          ),
          presentation_specs.ResourcePresentationSpec(
              '--backup-vault',
              GetBackupVaultResourceSpec(),
              """The backup vault where the backups gets stored using this
              backup plan.
              """,
              flag_name_overrides={
                  'location': '',
              },
              required=True,
          ),
      ],
      command_level_fallthroughs={
          '--backup-vault.location': ['BACKUP_PLAN.location'],
      },
  ).AddToParser(parser)


def AddEnforcedRetention(parser, required):
  """Adds a positional enforced-retention argument to parser.

  Args:
    parser: argparse.Parser: Parser object for command line inputs.
    required: Whether or not --backup-min-enforced-retention is required.
  """
  parser.add_argument(
      '--backup-min-enforced-retention',
      required=required,
      type=arg_parsers.Duration(
          lower_bound='0', upper_bound='36159d', parsed_unit='s'
      ),
      help=(
          'Backups will be kept for this minimum period before they can be'
          ' deleted. Once the effective time is reached, the enforced retention'
          ' period cannot be decreased or removed. The value must be specified'
          ' in relative time format (e.g. p1d, p1m, p1m1d).'
      ),
  )


def AddBackupRetentionInheritance(parser):
  """Adds the --backup-retention-inheritance flag to the given parser.

  Args:
    parser: argparse.Parser: Parser object for command line inputs.
  """
  parser.add_argument(
      '--backup-retention-inheritance',
      required=False,
      choices=[
          'inherit-vault-retention',
          'match-backup-expire-time',
      ],
      help=(
          'The inheritance mode for enforced retention end time of the backup'
          ' within this backup vault. Once set, the inheritance mode cannot be'
          ' changed. Default is inherit-vault-retention. If set to'
          ' inherit-vault-retention, the backup retention period will be'
          ' inherited from the backup vault. If set to'
          ' match-backup-expire-time, the backup retention period will  be the'
          ' same as the backup expiration time. '
      ),
  )


def AddUpdateBackupFlags(parser):
  """Adds update backup flags to the given parser."""
  group = parser.add_argument_group('Update Backup Flags')
  help_text = """
   Backups cannot be deleted until this time or later. This period can be extended, but not shortened.
   It should be specified in the format of "YYYY-MM-DD".

   * For backup configured with a backup appliance, there are additional restrictions:
     1. Enforced retention cannot be extended past the expiry time.
     2. Enforced retention can only be updated for finalized backups.
  """
  group.add_argument(
      '--enforced-retention-end-time',
      type=arg_parsers.Datetime.Parse,
      help=help_text,
  )
  group.add_argument(
      '--expire-time',
      type=arg_parsers.Datetime.Parse,
      help=(
          'The date when this backup is automatically expired. This date can'
          ' be extended, but not shortened. It should be specified in the'
          ' format of "YYYY-MM-DD".'
      ),
  )

  # A backup can only be updated with one of these flags at a time.
  label_update_group = parser.add_mutually_exclusive_group()

  label_update_group.add_argument(
      '--update-labels',
      metavar='KEY=VALUE',
      type=arg_parsers.ArgDict(value_type=str),
      action=arg_parsers.UpdateAction,
      help=textwrap.dedent("""\
      A list of labels to apply to the backup.

      Keys must start with a lowercase letter and contain only lowercase
      letters, numbers, and hyphens, and must be between 1 and 63 characters
      long. Values must contain only lowercase letters, numbers, and hyphens,
      and must be between 0 and 63 characters long.

      For example: `--update-labels=env=prod,team=storage`
      """))

  label_update_group.add_argument(
      '--remove-labels',
      metavar='KEY',
      type=arg_parsers.ArgList(min_length=1),
      help="""
      A list of label keys to remove from the backup.

      If a label does not exist, it is silently ignored.
      """)

  label_update_group.add_argument(
      '--clear-labels',
      action='store_true',
      help="""
      Remove all labels from the backup.

      If the backup has no labels, this operation is a no-op.
      """)


def AddOutputFormat(parser, output_format):
  parser.display_info.AddFormat(output_format)
  parser.display_info.AddTransforms({
      'backupMinimumEnforcedRetentionDuration': util.TransformEnforcedRetention,
  })


def AddDescription(parser, help_text=None):
  """Adds the --description flag to the given parser."""
  final_help_text = (
      help_text
      or 'Optional description for the backup vault (2048 characters or less).'
  )
  parser.add_argument('--description', type=str, help=final_help_text)


def AddLabels(parser, help_text=None):
  """Adds the --labels flag to the given parser."""
  final_help_text = (
      help_text
      or 'Optional resource labels to represent metadata provided by the user.'
  )
  parser.add_argument(
      '--labels',
      metavar='KEY=VALUE',
      type=arg_parsers.ArgDict(),
      action=arg_parsers.UpdateAction,
      help=final_help_text,
  )


def AddEffectiveTime(parser):
  """Adds the --effective-time flag to the given parser."""
  help_text = (
      'The time at which the enforced retention period becomes locked. This'
      ' flag is mutually exclusive with --unlock-backup-min-enforced-retention.'
  )
  parser.add_argument(
      '--effective-time',
      type=arg_parsers.Datetime.Parse,
      help=help_text,
  )


def AddAllowMissing(parser, resource):
  """Adds the --allow-missing flag to the given parser for delete operation to return success and perform no action when there is no matching resource."""
  help_text = (
      'Allow idempotent deletion of {resource}. The request will still succeed'
      ' in case the {resource} does not exist.'
  )
  parser.add_argument(
      '--allow-missing',
      action='store_true',
      help=help_text.format(resource=resource),
  )


def AddUnlockBackupMinEnforcedRetention(parser):
  """Adds the --unlock-backup-min-enforced-retention flag to the given parser."""
  help_text = (
      'Removes the lock on the backup minimum enforced retention period, and'
      ' resets the effective time. When unlocked, the enforced retention period'
      ' can be changed at any time. This flag is mutually exclusive with'
      ' --effective-time.'
  )
  parser.add_argument(
      '--unlock-backup-min-enforced-retention',
      action='store_true',
      help=help_text,
  )


def AddBackupVaultAccessRestrictionEnumFlag(parser, command: str):
  """Adds Backup Vault's Access Restriction flag to the parser."""
  choices = [
      'within-project',
      'within-org',
      'unrestricted',
      'within-org-but-unrestricted-for-ba',
  ]
  if command == 'create':
    help_text = (
        'Authorize certain sources and destinations for data being sent into,'
        ' or restored from, the backup vault being created. This choice '
        ' determines the type of resources that can be stored.'
        ' Restricting access to within your project or organization limits'
        ' the resources to those managed through the Google Cloud console'
        ' (e.g., Compute Engine VMs). Unrestricted access is required for'
        ' resources managed through the management console (e.g., VMware'
        ' Engine VMs, databases, and file systems).'
    )
    default = 'within-org'
  else:
    help_text = textwrap.dedent("""\
        Authorize certain sources and destinations for data being sent into, or restored from the current backup vault.

        Access restrictions can be modified to be more or less restrictive.

            ::: More restrictive access restriction update will fail by default if there will be non compliant Data Sources.
            To allow such updates, use the --force-update-access-restriction flag.
            :::  For Google Cloud Console resources, the following changes are allowed to make access restrictions more restrictive:
                *   `UNRESTRICTED` to `WITHIN_PROJECT` / `WITHIN_ORG_BUT_UNRESTRICTED_FOR_BA` / `WITHIN_ORGANIZATION`
                *   `WITHIN_PROJECT` to `WITHIN_ORGANIZATION` / `WITHIN_ORG_BUT_UNRESTRICTED_FOR_BA`

            ::: For Management Server resources, the following changes are allowed to make access restrictions more restrictive:
                *   `UNRESTRICTED` to `WITHIN_PROJECT` / `WITHIN_ORG_BUT_UNRESTRICTED_FOR_BA` / `WITHIN_ORGANIZATION`
                *   `WITHIN_PROJECT` to `WITHIN_ORGANIZATION` / `WITHIN_ORG_BUT_UNRESTRICTED_FOR_BA`

            :::   For both Google Cloud Console and Management Server resources, the following changes are allowed to make access restrictions more restrictive:
                *   `UNRESTRICTED` to `WITHIN_PROJECT` / `WITHIN_ORG_BUT_UNRESTRICTED_FOR_BA` / `WITHIN_ORGANIZATION`
                *   `WITHIN_PROJECT` to `WITHIN_ORGANIZATION` / `WITHIN_ORG_BUT_UNRESTRICTED_FOR_BA`

            ::: For Google Cloud Console resources,  the following changes are allowed to make access restrictions less restrictive:
                *   `WITHIN_ORGANIZATION` to `UNRESTRICTED` / `WITHIN_ORG_BUT_UNRESTRICTED_FOR_BA`
                *   `WITHIN_PROJECT` to `UNRESTRICTED`
                *   `WITHIN_ORG_BUT_UNRESTRICTED_FOR_BA` to `UNRESTRICTED`

            ::: For Management Server resources, the following changes are allowed to make access restrictions less restrictive:
                *   `WITHIN_ORG_BUT_UNRESTRICTED_FOR_BA` to `UNRESTRICTED`
        """)
    default = None

  parser.add_argument(
      '--access-restriction',
      choices=choices,
      default=default,
      help=help_text,
  )


def AddForceUpdateAccessRestriction(parser):
  """Adds the --force-update-access-restriction flag to the given parser."""
  help_text = (
      'If set, the access restriction can be updated even if there are'
      ' non-compliant data sources. Backups for those data sources will fail'
      ' afterward.'
  )
  parser.add_argument(
      '--force-update-access-restriction',
      action='store_true',
      help=help_text,
  )


def AddKmsKey(parser):
  """Adds the --kms-key flag to the given parser."""
  parser.add_argument(
      '--kms-key',
      required=False,
      type=str,
      help=(
          'The Cloud KMS key resource name to be used for encryption. Format:'
          ' projects/{project}/locations/{location}/keyRings/{ring}/cryptoKeys/{key}'
      ),
  )


def AddResourceType(parser, help_text):
  """Adds a positional resource-type argument to parser.

  Args:
    parser: argparse.Parser: Parser object for command line inputs.
    help_text: Help text for the resource-type argument.
  """
  parser.add_argument(
      '--resource-type',
      required=True,
      type=str,
      help=help_text,
  )


def AddLogRetentionDays(parser, hidden=True):
  """Adds a positional log-retention-days argument to parser.

  Args:
    parser: argparse.Parser: Parser object for command line inputs.
    hidden: Whether or not --log-retention-days is hidden. Default is True.
  """
  parser.add_argument(
      '--log-retention-days',
      required=False,
      hidden=hidden,
      type=int,
      help=textwrap.dedent("""\
          Configures how long logs will be stored. It is defined in "days".
          This value should be greater than or equal to minimum enforced log
          retention duration of the backup vault."""),
  )


def AddBackupRule(parser, required=False):
  """Adds a positional backup-rule argument to parser.

  Args:
    parser: argparse.Parser: Parser object for command line inputs.
    required: Whether or not --backup-rule is required.
  """

  rule_id_validator = arg_parsers.RegexpValidator(
      r'[a-z][a-z0-9-]{0,62}',
      'Invalid rule-id. This human-readable name must be unique and start with'
      ' a lowercase letter followed by up to 62 lowercase letters, numbers, or'
      ' hyphens',
  )

  week_day_of_month_options = [
      f'{week}-{day}'
      for week, day in itertools.product(WEEK_OPTIONS, DAY_OPTIONS.values())
  ]

  def ArgListParser(obj_parser: Any, delim: str = ' ') -> arg_parsers.ArgList:
    return arg_parsers.ArgList(obj_parser, custom_delim_char=delim)

  recurrence_validator = util.GetOneOfValidator(
      'recurrence', RECURRENCE_OPTIONS
  )

  parser.add_argument(
      '--backup-rule',
      required=required,
      type=arg_parsers.ArgDict(
          spec={
              'rule-id': rule_id_validator,
              'retention-days': int,
              'recurrence': recurrence_validator,
              'backup-window-start': arg_parsers.BoundedInt(0, 23),
              'backup-window-end': arg_parsers.BoundedInt(1, 24),
              'time-zone': str,
              'hourly-frequency': int,
              'days-of-week': ArgListParser(
                  util.OptionsMapValidator(DAY_OPTIONS).Parse
              ),
              'days-of-month': ArgListParser(arg_parsers.BoundedInt(1, 31)),
              'months': ArgListParser(
                  util.OptionsMapValidator(MONTH_OPTIONS).Parse
              ),
              'week-day-of-month': util.GetOneOfValidator(
                  'week-day-of-month', week_day_of_month_options
              ),
          },
          required_keys=[
              'rule-id',
              'recurrence',
              'retention-days',
              'backup-window-start',
              'backup-window-end',
          ],
      ),
      action='append',
      metavar='PROPERTY=VALUE',
      help=textwrap.dedent(f"""\
          Backup rule that defines parameters for when and how a backup
          is created. This flag can be repeated to create more backup rules.

          {BACKUP_RULE_COMMON_HELP_TEXT}
          """),
  )


def AddUpdateBackupRule(parser: argparse.ArgumentParser):
  """Adds a positional backup-rule argument to parser.

  Args:
    parser: argparse.Parser: Parser object for command line inputs.
  """

  rule_id_validator = arg_parsers.RegexpValidator(
      r'[a-z][a-z0-9-]{0,62}',
      'Invalid rule-id. This human-readable name must be unique and start with'
      ' a lowercase letter followed by up to 62 lowercase letters, numbers, or'
      ' hyphens',
  )

  week_day_of_month_options = [
      f'{week}-{day}'
      for week, day in itertools.product(WEEK_OPTIONS, DAY_OPTIONS.values())
  ]

  def ArgListParser(obj_parser: Any, delim: str = ' ') -> arg_parsers.ArgList:
    return arg_parsers.ArgList(obj_parser, custom_delim_char=delim)

  recurrence_validator = util.GetOneOfValidator(
      'recurrence', RECURRENCE_OPTIONS
  )

  parser.add_argument(
      '--backup-rule',
      type=arg_parsers.ArgDict(
          spec={
              'rule-id': rule_id_validator,
              'retention-days': int,
              'recurrence': recurrence_validator,
              'backup-window-start': arg_parsers.BoundedInt(0, 23),
              'backup-window-end': arg_parsers.BoundedInt(1, 24),
              'time-zone': str,
              'hourly-frequency': int,
              'days-of-week': ArgListParser(
                  util.OptionsMapValidator(DAY_OPTIONS).Parse
              ),
              'days-of-month': ArgListParser(arg_parsers.BoundedInt(1, 31)),
              'months': ArgListParser(
                  util.OptionsMapValidator(MONTH_OPTIONS).Parse
              ),
              'week-day-of-month': util.GetOneOfValidator(
                  'week-day-of-month', week_day_of_month_options
              ),
          },
          required_keys=['rule-id'],
      ),
      action='append',
      metavar='PROPERTY=VALUE',
      help=textwrap.dedent(f"""\
          Full definition of an existing backup rule with updated values.
          The existing backup rule is replaced by this new set of values.
          This flag can be repeated to update multiple backup rules.
          It is not allowed to pass the same rule-id in this flag more than once
          in the same command.

          {BACKUP_RULE_COMMON_HELP_TEXT}
          """),
  )


def AddAddBackupRule(parser):
  """Adds flags required to add a backup rule.

  Args:
    parser: argparse.Parser: Parser object for command line inputs.
  """

  rule_id_validator = arg_parsers.RegexpValidator(
      r'[a-z][a-z0-9-]{0,62}',
      'Invalid rule-id. This human-readable name must be unique and start with'
      ' a lowercase letter followed by up to 62 lowercase letters, numbers, or'
      ' hyphens',
  )

  week_day_of_month_options = [
      f'{week}-{day}'
      for week, day in itertools.product(WEEK_OPTIONS, DAY_OPTIONS.values())
  ]

  def ArgListParser(obj_parser: Any, delim: str = ' ') -> arg_parsers.ArgList:
    return arg_parsers.ArgList(obj_parser, custom_delim_char=delim)

  recurrence_validator = util.GetOneOfValidator(
      'recurrence', RECURRENCE_OPTIONS
  )

  parser.add_argument(
      '--add-backup-rule',
      required=False,
      type=arg_parsers.ArgDict(
          spec={
              'rule-id': rule_id_validator,
              'retention-days': int,
              'recurrence': recurrence_validator,
              'backup-window-start': arg_parsers.BoundedInt(0, 23),
              'backup-window-end': arg_parsers.BoundedInt(1, 24),
              'time-zone': str,
              'hourly-frequency': int,
              'days-of-week': ArgListParser(
                  util.OptionsMapValidator(DAY_OPTIONS).Parse
              ),
              'days-of-month': ArgListParser(arg_parsers.BoundedInt(1, 31)),
              'months': ArgListParser(
                  util.OptionsMapValidator(MONTH_OPTIONS).Parse
              ),
              'week-day-of-month': util.GetOneOfValidator(
                  'week-day-of-month', week_day_of_month_options
              ),
          },
          required_keys=[
              'rule-id',
              'recurrence',
              'retention-days',
              'backup-window-start',
              'backup-window-end',
          ],
      ),
      action='append',
      metavar='PROPERTY=VALUE',
      help=textwrap.dedent("""\
          Parameters of backup rule to be added to the Backup Plan. This flag can be repeated to add more backup rules.
          """),
  )


def AddRemoveBackupRule(parser):
  """Adds flags required to remove a backup rule.

  Args:
    parser: argparse.Parser: Parser object for command line inputs.
  """
  parser.add_argument(
      '--remove-backup-rule',
      required=False,
      type=str,
      help=textwrap.dedent("""\
          Name of an existing backup rule to be removed from the Backup Plan. This flag can be repeated to remove more backup rules.
          """),
      action='append',
      metavar='RULE-ID',
  )


def AddBackupRulesFromFile(parser):
  """Adds flags required to add backup rules from a file.

  Args:
    parser: argparse.Parser: Parser object for command line inputs.
  """
  parser.add_argument(
      '--backup-rules-from-file',
      required=False,
      type=arg_parsers.FileContents(),
      help='Path to a YAML or JSON file containing backup rules.',
  )


def AddCloudsqlEdition(parser, help_text):
  """Adds a positional edition argument to parser.

  Args:
    parser: argparse.Parser: Parser object for command line inputs.
    help_text: Help text for the edition argument.
  """
  parser.add_argument(
      '--cloudsql-edition',
      required=False,
      type=str,
      help=help_text,
  )


def AddMaxCustomOnDemandRetentionDays(parser):
  """Adds a max-custom-on-demand-retention-days argument to parser."""
  parser.add_argument(
      '--max-custom-on-demand-retention-days',
      required=False,
      type=int,
      help=textwrap.dedent("""\
          Configure the maximum retention period for on-demand backups.
          The value must be greater than or equal to the minimum enforced
          retention period set on the backup vault."""),
  )


def AddDiskBackupPlanProperties(parser):
  """Adds the --disk-properties flag to the given parser."""
  help_text = textwrap.dedent("""\
      Workload-specific properties for disk backups generated by this plan.

      *guest-flush*::: Indicates whether to perform a guest flush operation
      before taking a disk backup. When set to true, the system will attempt
      to ensure application-consistent backups. When set to false, the system
      will create crash-consistent backups. Default value is false.

      Example: --disk-properties=guest-flush=true
      """)
  parser.add_argument(
      '--disk-properties',
      metavar='PROPERTY=VALUE',
      type=arg_parsers.ArgDict(
          spec={
              'guest-flush': arg_parsers.ArgBoolean(),
          },
      ),
      help=help_text,
  )


def AddComputeInstanceBackupPlanProperties(parser):
  """Adds the --compute-instance-properties flag to the given parser."""
  help_text = textwrap.dedent("""\
      Workload-specific properties for Compute Engine instance backups generated by this plan.

      *guest-flush*::: Indicates whether to perform a guest flush operation
      before taking a compute backup. When set to true, the system will attempt
      to ensure application-consistent backups. When set to false, the system
      will create crash-consistent backups. Default value is false.

      Example: --compute-instance-properties=guest-flush=true
      """)
  parser.add_argument(
      '--compute-instance-properties',
      metavar='PROPERTY=VALUE',
      type=arg_parsers.ArgDict(
          spec={
              'guest-flush': arg_parsers.ArgBoolean(),
          },
      ),
      help=help_text,
  )


def AddUseProjectServiceAccountFlag(parser):
  """Adds the --use-project-service-account flag to the given parser."""
  parser.add_argument(
      '--use-project-service-account',
      action='store_true',
      required=False,
      default=False,
      help=(
          'If set, the BackupDR Project Service Agent (P4SA) credentials '
          'will be used for the restore operation instead of the Backup Vault '
          'Service Account. This is only supported for same-project restores.'
      ),
  )



# -*- coding: utf-8 -*- #
# Copyright 2025 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Command for deleting snapshot groups."""

from googlecloudsdk.api_lib.compute import base_classes
from googlecloudsdk.api_lib.compute import utils
from googlecloudsdk.calliope import base
from googlecloudsdk.command_lib.compute import flags as compute_flags
from googlecloudsdk.command_lib.compute import scope as compute_scope
from googlecloudsdk.command_lib.compute.snapshot_groups import flags

DELETE = 'Delete'

DETAILED_HELP = {
    'EXAMPLES':
        """\
        To delete Compute Engine snapshot groups with the names 'sg-1'
        and 'sg-2', run:

          $ {command} sg-1 sg-2
        """,
}


def _CommonArgs(parser):
  """A helper function to build args for all API versions."""
  flags.MakeSnapshotGroupArg(plural=True).AddArgument(
      parser, operation_type='delete'
  )


def _RunDelete(self, args):
  """Shared logic for deleting snapshot groups."""
  holder = base_classes.ComputeApiHolder(self.ReleaseTrack())
  client = holder.client

  snapshot_group_refs = flags.MakeSnapshotGroupArg(
      plural=True
  ).ResolveAsResource(
      args,
      holder.resources,
      scope_lister=compute_flags.GetDefaultScopeLister(client),
      default_scope=compute_scope.ScopeEnum.GLOBAL,
  )

  utils.PromptForDeletion(snapshot_group_refs)
  requests = []
  for snapshot_group_ref in snapshot_group_refs:
    requests.append((
        client.apitools_client.snapshotGroups,
        DELETE,
        client.messages.ComputeSnapshotGroupsDeleteRequest(
            project=snapshot_group_ref.project,
            snapshotGroup=snapshot_group_ref.snapshotGroup,
        ),
    ))

  return client.MakeRequests(requests)


@base.ReleaseTracks(base.ReleaseTrack.ALPHA)
@base.DefaultUniverseOnly
class DeleteAlpha(base.DeleteCommand):
  """Delete Compute Engine snapshot groups.

  *{command}* deletes one or more Compute Engine snapshot groups.
  """

  detailed_help = DETAILED_HELP

  @staticmethod
  def Args(parser):
    _CommonArgs(parser)

  def Run(self, args):
    return _RunDelete(self, args)


@base.Hidden  # Hide this command from public documentation
@base.ReleaseTracks(base.ReleaseTrack.BETA)
@base.DefaultUniverseOnly
class DeleteBeta(base.DeleteCommand):
  """Delete Compute Engine snapshot groups.

  *{command}* deletes one or more Compute Engine snapshot groups.
  """

  detailed_help = DETAILED_HELP

  @staticmethod
  def Args(parser):
    _CommonArgs(parser)

  def Run(self, args):
    return _RunDelete(self, args)




"""Generated client library for stream version v1alpha1."""
# NOTE: This file is autogenerated and should not be edited by hand.

from apitools.base.py import base_api
from googlecloudsdk.generated_clients.apis.stream.v1alpha1 import stream_v1alpha1_messages as messages


class StreamV1alpha1(base_api.BaseApiClient):
  """Generated client library for service stream version v1alpha1."""

  MESSAGES_MODULE = messages
  BASE_URL = 'https://stream.googleapis.com/'
  MTLS_BASE_URL = 'https://stream.mtls.googleapis.com/'

  _PACKAGE = 'stream'
  _SCOPES = ['https://www.googleapis.com/auth/cloud-platform']
  _VERSION = 'v1alpha1'
  _CLIENT_ID = 'CLIENT_ID'
  _CLIENT_SECRET = 'CLIENT_SECRET'
  _USER_AGENT = 'google-cloud-sdk'
  _CLIENT_CLASS_NAME = 'StreamV1alpha1'
  _URL_VERSION = 'v1alpha1'
  _API_KEY = None

  def __init__(self, url='', credentials=None,
               get_credentials=True, http=None, model=None,
               log_request=False, log_response=False,
               credentials_args=None, default_global_params=None,
               additional_http_headers=None, response_encoding=None):
    """Create a new stream handle."""
    url = url or self.BASE_URL
    super(StreamV1alpha1, self).__init__(
        url, credentials=credentials,
        get_credentials=get_credentials, http=http, model=model,
        log_request=log_request, log_response=log_response,
        credentials_args=credentials_args,
        default_global_params=default_global_params,
        additional_http_headers=additional_http_headers,
        response_encoding=response_encoding)
    self.projects_locations_operations = self.ProjectsLocationsOperationsService(self)
    self.projects_locations_streamContents = self.ProjectsLocationsStreamContentsService(self)
    self.projects_locations_streamInstances = self.ProjectsLocationsStreamInstancesService(self)
    self.projects_locations = self.ProjectsLocationsService(self)
    self.projects = self.ProjectsService(self)

  class ProjectsLocationsOperationsService(base_api.BaseApiService):
    """Service class for the projects_locations_operations resource."""

    _NAME = 'projects_locations_operations'

    def __init__(self, client):
      super(StreamV1alpha1.ProjectsLocationsOperationsService, self).__init__(client)
      self._upload_configs = {
          }

    def Cancel(self, request, global_params=None):
      r"""Starts asynchronous cancellation on a long-running operation. The server makes a best effort to cancel the operation, but success is not guaranteed. If the server doesn't support this method, it returns `google.rpc.Code.UNIMPLEMENTED`. Clients can use Operations.GetOperation or other methods to check whether the cancellation succeeded or whether the operation completed despite cancellation. On successful cancellation, the operation is not deleted; instead, it becomes an operation with an Operation.error value with a google.rpc.Status.code of `1`, corresponding to `Code.CANCELLED`.

      Args:
        request: (StreamProjectsLocationsOperationsCancelRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (Empty) The response message.
      """
      config = self.GetMethodConfig('Cancel')
      return self._RunMethod(
          config, request, global_params=global_params)

    Cancel.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='v1alpha1/projects/{projectsId}/locations/{locationsId}/operations/{operationsId}:cancel',
        http_method='POST',
        method_id='stream.projects.locations.operations.cancel',
        ordered_params=['name'],
        path_params=['name'],
        query_params=[],
        relative_path='v1alpha1/{+name}:cancel',
        request_field='cancelOperationRequest',
        request_type_name='StreamProjectsLocationsOperationsCancelRequest',
        response_type_name='Empty',
        supports_download=False,
    )

    def Delete(self, request, global_params=None):
      r"""Deletes a long-running operation. This method indicates that the client is no longer interested in the operation result. It does not cancel the operation. If the server doesn't support this method, it returns `google.rpc.Code.UNIMPLEMENTED`.

      Args:
        request: (StreamProjectsLocationsOperationsDeleteRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (Empty) The response message.
      """
      config = self.GetMethodConfig('Delete')
      return self._RunMethod(
          config, request, global_params=global_params)

    Delete.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='v1alpha1/projects/{projectsId}/locations/{locationsId}/operations/{operationsId}',
        http_method='DELETE',
        method_id='stream.projects.locations.operations.delete',
        ordered_params=['name'],
        path_params=['name'],
        query_params=[],
        relative_path='v1alpha1/{+name}',
        request_field='',
        request_type_name='StreamProjectsLocationsOperationsDeleteRequest',
        response_type_name='Empty',
        supports_download=False,
    )

    def Get(self, request, global_params=None):
      r"""Gets the latest state of a long-running operation. Clients can use this method to poll the operation result at intervals as recommended by the API service.

      Args:
        request: (StreamProjectsLocationsOperationsGetRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (Operation) The response message.
      """
      config = self.GetMethodConfig('Get')
      return self._RunMethod(
          config, request, global_params=global_params)

    Get.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='v1alpha1/projects/{projectsId}/locations/{locationsId}/operations/{operationsId}',
        http_method='GET',
        method_id='stream.projects.locations.operations.get',
        ordered_params=['name'],
        path_params=['name'],
        query_params=[],
        relative_path='v1alpha1/{+name}',
        request_field='',
        request_type_name='StreamProjectsLocationsOperationsGetRequest',
        response_type_name='Operation',
        supports_download=False,
    )

    def List(self, request, global_params=None):
      r"""Lists operations that match the specified filter in the request. If the server doesn't support this method, it returns `UNIMPLEMENTED`.

      Args:
        request: (StreamProjectsLocationsOperationsListRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (ListOperationsResponse) The response message.
      """
      config = self.GetMethodConfig('List')
      return self._RunMethod(
          config, request, global_params=global_params)

    List.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='v1alpha1/projects/{projectsId}/locations/{locationsId}/operations',
        http_method='GET',
        method_id='stream.projects.locations.operations.list',
        ordered_params=['name'],
        path_params=['name'],
        query_params=['filter', 'pageSize', 'pageToken', 'returnPartialSuccess'],
        relative_path='v1alpha1/{+name}/operations',
        request_field='',
        request_type_name='StreamProjectsLocationsOperationsListRequest',
        response_type_name='ListOperationsResponse',
        supports_download=False,
    )

  class ProjectsLocationsStreamContentsService(base_api.BaseApiService):
    """Service class for the projects_locations_streamContents resource."""

    _NAME = 'projects_locations_streamContents'

    def __init__(self, client):
      super(StreamV1alpha1.ProjectsLocationsStreamContentsService, self).__init__(client)
      self._upload_configs = {
          }

    def Build(self, request, global_params=None):
      r"""Builds the content to a Stream compatible format using the associated sources in a consumer cloud storage bucket. A new content version is created with the user-specified tag if the build succeeds. The returned Operation can be used to track the build status by polling operations.get. The Operation will complete when the build is done. Returns [StreamContent] in the Operation.response field on successful completion.

      Args:
        request: (StreamProjectsLocationsStreamContentsBuildRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (Operation) The response message.
      """
      config = self.GetMethodConfig('Build')
      return self._RunMethod(
          config, request, global_params=global_params)

    Build.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='v1alpha1/projects/{projectsId}/locations/{locationsId}/streamContents/{streamContentsId}:build',
        http_method='POST',
        method_id='stream.projects.locations.streamContents.build',
        ordered_params=['name'],
        path_params=['name'],
        query_params=[],
        relative_path='v1alpha1/{+name}:build',
        request_field='buildStreamContentRequest',
        request_type_name='StreamProjectsLocationsStreamContentsBuildRequest',
        response_type_name='Operation',
        supports_download=False,
    )

    def Create(self, request, global_params=None):
      r"""Creates a new StreamContent that manages the metadata and builds of user-provided Stream compatible content sources in a consumer cloud storage bucket. The returned Operation can be used to track the creation status by polling operations.get. The Operation will complete when the creation is done. Returns [StreamContent] in the Operation.response field on successful completion.

      Args:
        request: (StreamProjectsLocationsStreamContentsCreateRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (Operation) The response message.
      """
      config = self.GetMethodConfig('Create')
      return self._RunMethod(
          config, request, global_params=global_params)

    Create.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='v1alpha1/projects/{projectsId}/locations/{locationsId}/streamContents',
        http_method='POST',
        method_id='stream.projects.locations.streamContents.create',
        ordered_params=['parent'],
        path_params=['parent'],
        query_params=['requestId', 'streamContentId'],
        relative_path='v1alpha1/{+parent}/streamContents',
        request_field='streamContent',
        request_type_name='StreamProjectsLocationsStreamContentsCreateRequest',
        response_type_name='Operation',
        supports_download=False,
    )

    def Delete(self, request, global_params=None):
      r"""Deletes a single StreamContent. This method removes the version history of content builds but does not delete any content source in the consumer cloud storage bucket. The returned Operation can be used to track the deletion status by polling operations.get. The Operation will complete when the deletion is done. Returns Empty in the Operation.response field on successful completion.

      Args:
        request: (StreamProjectsLocationsStreamContentsDeleteRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (Operation) The response message.
      """
      config = self.GetMethodConfig('Delete')
      return self._RunMethod(
          config, request, global_params=global_params)

    Delete.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='v1alpha1/projects/{projectsId}/locations/{locationsId}/streamContents/{streamContentsId}',
        http_method='DELETE',
        method_id='stream.projects.locations.streamContents.delete',
        ordered_params=['name'],
        path_params=['name'],
        query_params=['requestId'],
        relative_path='v1alpha1/{+name}',
        request_field='',
        request_type_name='StreamProjectsLocationsStreamContentsDeleteRequest',
        response_type_name='Operation',
        supports_download=False,
    )

    def Get(self, request, global_params=None):
      r"""Gets details of a single StreamContent.

      Args:
        request: (StreamProjectsLocationsStreamContentsGetRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (StreamContent) The response message.
      """
      config = self.GetMethodConfig('Get')
      return self._RunMethod(
          config, request, global_params=global_params)

    Get.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='v1alpha1/projects/{projectsId}/locations/{locationsId}/streamContents/{streamContentsId}',
        http_method='GET',
        method_id='stream.projects.locations.streamContents.get',
        ordered_params=['name'],
        path_params=['name'],
        query_params=[],
        relative_path='v1alpha1/{+name}',
        request_field='',
        request_type_name='StreamProjectsLocationsStreamContentsGetRequest',
        response_type_name='StreamContent',
        supports_download=False,
    )

    def List(self, request, global_params=None):
      r"""Lists StreamContents in a given project and location.

      Args:
        request: (StreamProjectsLocationsStreamContentsListRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (ListStreamContentsResponse) The response message.
      """
      config = self.GetMethodConfig('List')
      return self._RunMethod(
          config, request, global_params=global_params)

    List.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='v1alpha1/projects/{projectsId}/locations/{locationsId}/streamContents',
        http_method='GET',
        method_id='stream.projects.locations.streamContents.list',
        ordered_params=['parent'],
        path_params=['parent'],
        query_params=['filter', 'orderBy', 'pageSize', 'pageToken'],
        relative_path='v1alpha1/{+parent}/streamContents',
        request_field='',
        request_type_name='StreamProjectsLocationsStreamContentsListRequest',
        response_type_name='ListStreamContentsResponse',
        supports_download=False,
    )

    def Patch(self, request, global_params=None):
      r"""Updates the parameters of a single StreamContent.

      Args:
        request: (StreamProjectsLocationsStreamContentsPatchRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (Operation) The response message.
      """
      config = self.GetMethodConfig('Patch')
      return self._RunMethod(
          config, request, global_params=global_params)

    Patch.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='v1alpha1/projects/{projectsId}/locations/{locationsId}/streamContents/{streamContentsId}',
        http_method='PATCH',
        method_id='stream.projects.locations.streamContents.patch',
        ordered_params=['name'],
        path_params=['name'],
        query_params=['requestId', 'updateMask'],
        relative_path='v1alpha1/{+name}',
        request_field='streamContent',
        request_type_name='StreamProjectsLocationsStreamContentsPatchRequest',
        response_type_name='Operation',
        supports_download=False,
    )

  class ProjectsLocationsStreamInstancesService(base_api.BaseApiService):
    """Service class for the projects_locations_streamInstances resource."""

    _NAME = 'projects_locations_streamInstances'

    def __init__(self, client):
      super(StreamV1alpha1.ProjectsLocationsStreamInstancesService, self).__init__(client)
      self._upload_configs = {
          }

    def Create(self, request, global_params=None):
      r"""Creates a new StreamInstance that manages the turnup and rollout of the streaming service for a given StreamContent. The returned Operation can be used to track the creation status by polling operations.get. The Operation will complete when the creation is done. Returns [StreamInstance] in the Operation.response field on successful completion.

      Args:
        request: (StreamProjectsLocationsStreamInstancesCreateRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (Operation) The response message.
      """
      config = self.GetMethodConfig('Create')
      return self._RunMethod(
          config, request, global_params=global_params)

    Create.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='v1alpha1/projects/{projectsId}/locations/{locationsId}/streamInstances',
        http_method='POST',
        method_id='stream.projects.locations.streamInstances.create',
        ordered_params=['parent'],
        path_params=['parent'],
        query_params=['requestId', 'streamInstanceId'],
        relative_path='v1alpha1/{+parent}/streamInstances',
        request_field='streamInstance',
        request_type_name='StreamProjectsLocationsStreamInstancesCreateRequest',
        response_type_name='Operation',
        supports_download=False,
    )

    def Delete(self, request, global_params=None):
      r"""Deletes a single StreamInstance. This method tears down the streaming service of the associated StreamContent. The returned Operation can be used to track the deletion status by polling operations.get. The Operation will complete when the deletion is done. Returns Empty in the Operation.response field on successful completion.

      Args:
        request: (StreamProjectsLocationsStreamInstancesDeleteRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (Operation) The response message.
      """
      config = self.GetMethodConfig('Delete')
      return self._RunMethod(
          config, request, global_params=global_params)

    Delete.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='v1alpha1/projects/{projectsId}/locations/{locationsId}/streamInstances/{streamInstancesId}',
        http_method='DELETE',
        method_id='stream.projects.locations.streamInstances.delete',
        ordered_params=['name'],
        path_params=['name'],
        query_params=['requestId'],
        relative_path='v1alpha1/{+name}',
        request_field='',
        request_type_name='StreamProjectsLocationsStreamInstancesDeleteRequest',
        response_type_name='Operation',
        supports_download=False,
    )

    def Get(self, request, global_params=None):
      r"""Gets details of a single StreamInstance.

      Args:
        request: (StreamProjectsLocationsStreamInstancesGetRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (StreamInstance) The response message.
      """
      config = self.GetMethodConfig('Get')
      return self._RunMethod(
          config, request, global_params=global_params)

    Get.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='v1alpha1/projects/{projectsId}/locations/{locationsId}/streamInstances/{streamInstancesId}',
        http_method='GET',
        method_id='stream.projects.locations.streamInstances.get',
        ordered_params=['name'],
        path_params=['name'],
        query_params=[],
        relative_path='v1alpha1/{+name}',
        request_field='',
        request_type_name='StreamProjectsLocationsStreamInstancesGetRequest',
        response_type_name='StreamInstance',
        supports_download=False,
    )

    def List(self, request, global_params=None):
      r"""Lists StreamInstances in a given project and location.

      Args:
        request: (StreamProjectsLocationsStreamInstancesListRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (ListStreamInstancesResponse) The response message.
      """
      config = self.GetMethodConfig('List')
      return self._RunMethod(
          config, request, global_params=global_params)

    List.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='v1alpha1/projects/{projectsId}/locations/{locationsId}/streamInstances',
        http_method='GET',
        method_id='stream.projects.locations.streamInstances.list',
        ordered_params=['parent'],
        path_params=['parent'],
        query_params=['filter', 'orderBy', 'pageSize', 'pageToken'],
        relative_path='v1alpha1/{+parent}/streamInstances',
        request_field='',
        request_type_name='StreamProjectsLocationsStreamInstancesListRequest',
        response_type_name='ListStreamInstancesResponse',
        supports_download=False,
    )

    def Patch(self, request, global_params=None):
      r"""Updates the parameters of a single StreamInstance.

      Args:
        request: (StreamProjectsLocationsStreamInstancesPatchRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (Operation) The response message.
      """
      config = self.GetMethodConfig('Patch')
      return self._RunMethod(
          config, request, global_params=global_params)

    Patch.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='v1alpha1/projects/{projectsId}/locations/{locationsId}/streamInstances/{streamInstancesId}',
        http_method='PATCH',
        method_id='stream.projects.locations.streamInstances.patch',
        ordered_params=['name'],
        path_params=['name'],
        query_params=['requestId', 'updateMask'],
        relative_path='v1alpha1/{+name}',
        request_field='streamInstance',
        request_type_name='StreamProjectsLocationsStreamInstancesPatchRequest',
        response_type_name='Operation',
        supports_download=False,
    )

  class ProjectsLocationsService(base_api.BaseApiService):
    """Service class for the projects_locations resource."""

    _NAME = 'projects_locations'

    def __init__(self, client):
      super(StreamV1alpha1.ProjectsLocationsService, self).__init__(client)
      self._upload_configs = {
          }

    def Get(self, request, global_params=None):
      r"""Gets information about a location.

      Args:
        request: (StreamProjectsLocationsGetRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (Location) The response message.
      """
      config = self.GetMethodConfig('Get')
      return self._RunMethod(
          config, request, global_params=global_params)

    Get.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='v1alpha1/projects/{projectsId}/locations/{locationsId}',
        http_method='GET',
        method_id='stream.projects.locations.get',
        ordered_params=['name'],
        path_params=['name'],
        query_params=[],
        relative_path='v1alpha1/{+name}',
        request_field='',
        request_type_name='StreamProjectsLocationsGetRequest',
        response_type_name='Location',
        supports_download=False,
    )

    def List(self, request, global_params=None):
      r"""Lists information about the supported locations for this service. This method lists locations based on the resource scope provided in the ListLocationsRequest.name field: * **Global locations**: If `name` is empty, the method lists the public locations available to all projects. * **Project-specific locations**: If `name` follows the format `projects/{project}`, the method lists locations visible to that specific project. This includes public, private, or other project-specific locations enabled for the project. For gRPC and client library implementations, the resource name is passed as the `name` field. For direct service calls, the resource name is incorporated into the request path based on the specific service implementation and version.

      Args:
        request: (StreamProjectsLocationsListRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (ListLocationsResponse) The response message.
      """
      config = self.GetMethodConfig('List')
      return self._RunMethod(
          config, request, global_params=global_params)

    List.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='v1alpha1/projects/{projectsId}/locations',
        http_method='GET',
        method_id='stream.projects.locations.list',
        ordered_params=['name'],
        path_params=['name'],
        query_params=['extraLocationTypes', 'filter', 'includeUnrevealedLocations', 'pageSize', 'pageToken'],
        relative_path='v1alpha1/{+name}/locations',
        request_field='',
        request_type_name='StreamProjectsLocationsListRequest',
        response_type_name='ListLocationsResponse',
        supports_download=False,
    )

  class ProjectsService(base_api.BaseApiService):
    """Service class for the projects resource."""

    _NAME = 'projects'

    def __init__(self, client):
      super(StreamV1alpha1.ProjectsService, self).__init__(client)
      self._upload_configs = {
          }



# -*- coding: utf-8 -*- #
#
# Copyright 2025 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Common utility functions for Developer Connect Insights Configs."""

import datetime
from typing import Any, Dict, List, Optional, Set, Tuple

from apitools.base.py import exceptions as apitools_exceptions
from googlecloudsdk.api_lib.cloudresourcemanager import projects_api
from googlecloudsdk.api_lib.developer_connect import common
from googlecloudsdk.api_lib.developer_connect.insights_configs import discover_apphub
from googlecloudsdk.api_lib.developer_connect.insights_configs import discover_artifact_configs as discover_artifacts
from googlecloudsdk.api_lib.resource_manager import folders
from googlecloudsdk.api_lib.services import serviceusage
from googlecloudsdk.api_lib.util import apis
from googlecloudsdk.api_lib.util import waiter
from googlecloudsdk.calliope import base
from googlecloudsdk.command_lib.developer_connect import name
from googlecloudsdk.command_lib.iam import iam_util
from googlecloudsdk.command_lib.projects import util as projects_util
from googlecloudsdk.core import exceptions
from googlecloudsdk.core import log
from googlecloudsdk.core import resources
from googlecloudsdk.core.console import console_io


_ROLES = (
    'roles/developerconnect.insightsAgent',
)
_EXCLUDED_PROJECTS = ['cloudrun']
# Wait till service account is available for setIamPolicy
_MAX_WAIT_TIME_IN_MS = 20 * 1000
VERSION_MAP = {
    base.ReleaseTrack.ALPHA: 'v1',
    base.ReleaseTrack.BETA: 'v1',
    base.ReleaseTrack.GA: 'v1',
}


class InsightsConfigInitializationError(exceptions.InternalError):
  """Error initializing the Developer Connect Insights Config."""


def _get_p4sa(project, service_name):
  """Gets the P4SA for the given project and location.

  If the P4SA does not exist for this project, it will be created. Otherwise,
  the email address of the existing P4SA will be returned.

  Args:
    project: The project to get the P4SA for.
    service_name: The service name to get the P4SA for.

  Returns:
    The email address of the P4SA.
  """
  response = serviceusage.GenerateServiceIdentity(project, service_name)
  return response['email']


def _should_retry_http_error(
    exc_type, unused_exc_value, unused_exc_traceback, unused_state
):
  """Whether to retry the request when receiving errors.

  Args:
    exc_type: type of the raised exception.
    unused_exc_value: the instance of the raise the exception.
    unused_exc_traceback: Traceback, traceback encapsulating the call stack at
      the point where the exception occurred.
    unused_state: RetryerState, state of the retryer.

  Returns:
    True if exception and is due to NOT_FOUND or INVALID_ARGUMENT.
  """
  return (exc_type == apitools_exceptions.HttpBadRequestError or
          exc_type == apitools_exceptions.HttpNotFoundError)


# The messages module can also be accessed from client.MESSAGES_MODULE
def get_messages_module(release_track=base.ReleaseTrack.ALPHA):
  api_version = VERSION_MAP.get(release_track)
  return apis.GetMessagesModule('developerconnect', api_version)


class InsightsConfigClient(object):
  """Wrapper for Developer Connect Insights API client."""

  def __init__(self, release_track):
    api_version = VERSION_MAP.get(release_track)
    self.release_track = release_track
    self.client = apis.GetClientInstance('developerconnect', api_version)
    self._resource_parser = resources.Registry()
    self._resource_parser.RegisterApiByName('developerconnect', 'v1')
    self.messages = get_messages_module(release_track)
    self.api_version = api_version
    self.p4sa_email = None

  def create(
      self,
      insight_config_ref,
      app_hub,
      target_projects,
      user_artifact_configs,
      source_config,
  ):
    """Creates the insight config."""
    if app_hub:
      create_request = self.create_apphub_insights_config_request(
          insight_config_ref, app_hub, user_artifact_configs, source_config,
      )
    elif target_projects:
      create_request = self.create_project_scope_insights_config_request(
          insight_config_ref, target_projects, user_artifact_configs,
          source_config,
      )
    elif source_config:
      create_request = self.create_source_config_insights_config_request(
          insight_config_ref, user_artifact_configs, source_config
      )
    else:
      raise exceptions.Error(
          'You must specify a runtime context (--app-hub-application or '
          '--target-projects) and/or a source configuration (--source-config).'
      )

    try:
      return self.client.projects_locations_insightsConfigs.Create(
          request=create_request
      )
    except apitools_exceptions.HttpConflictError:
      raise exceptions.Error(
          f'Insights Config [{insight_config_ref.insightsConfigsId}] already'
          f' exists in project [{insight_config_ref.projectsId}] location'
          f' [{insight_config_ref.locationsId}].'
      )

  def create_apphub_insights_config_request(
      self,
      insight_config_ref,
      app_hub,
      user_artifact_configs,
      source_config,
  ) -> (
      get_messages_module().DeveloperconnectProjectsLocationsInsightsConfigsCreateRequest
  ):
    """Creates the insight config request for app hub application scope."""
    app_hub_application = name.parse_app_hub_application_uri(app_hub)
    # Check if the app hub application project is the same as the project
    # where the insight config is being created.
    is_project_id_mismatch = (
        app_hub_application.project_id != insight_config_ref.projectsId
    )
    is_project_num_mismatch = (
        str(app_hub_application.project_number) != insight_config_ref.projectsId
    )
    if is_project_id_mismatch and is_project_num_mismatch:
      raise InsightsConfigInitializationError(
          'Mismatch: App Hub application project'
          f' [{app_hub_application.project_id}] must be the same as the'
          ' project where the insight config is being created'
          f' [{insight_config_ref.projectsId}].'
      )

    # Handle the management project and get the dependent projects and gke
    # workloads.
    dependent_projects, gke_workloads, cloud_run_services = (
        self.find_apphub_resources_and_grant_sa_permissions(
            insight_config_ref, app_hub_application
        )
    )
    cais_artifact_configs_dict = self.get_artifact_configs_from_cais(
        gke_workloads, cloud_run_services
    )

    user_artifact_configs_dict = name.parse_artifact_configs(
        user_artifact_configs
    )
    merged_artifact_configs_dict = self.merge_artifact_configs(
        cais_artifact_configs_dict, user_artifact_configs_dict
    )

    # Get the artifact configs and add new projects to the dependent projects
    # set.
    artifact_projects, artifact_configs = self.build_artifact_configs(
        merged_artifact_configs_dict, cais_artifact_configs_dict
    )
    # Add the artifact projects to the dependent projects set.
    dependent_projects.update(artifact_projects)
    # Get the P4SA and grant IAM roles to it.
    if dependent_projects:
      # Sort projects for deterministic behavior, important for tests.
      self.init_service_account(
          insight_config_ref.projectsId,
          sorted(dependent_projects),
          management_project=False,
      )
    if source_config:
      return self.messages.DeveloperconnectProjectsLocationsInsightsConfigsCreateRequest(
          parent=insight_config_ref.Parent().RelativeName(),
          insightsConfigId=insight_config_ref.insightsConfigsId,
          insightsConfig=self.messages.InsightsConfig(
              name=insight_config_ref.RelativeName(),
              appHubApplication=app_hub_application.resource_name(),
              artifactConfigs=artifact_configs,
              sourceConfigs=[self.messages.SourceConfig(
                  gitRepositoryLink=source_config.get('git-repository-link'),
              )],
          ),
      )
    else:
      return self.messages.DeveloperconnectProjectsLocationsInsightsConfigsCreateRequest(
          parent=insight_config_ref.Parent().RelativeName(),
          insightsConfigId=insight_config_ref.insightsConfigsId,
          insightsConfig=self.messages.InsightsConfig(
              name=insight_config_ref.RelativeName(),
              appHubApplication=app_hub_application.resource_name(),
              artifactConfigs=artifact_configs,
          ),
      )

  def create_project_scope_insights_config_request(
      self,
      insight_config_ref,
      target_projects,
      user_artifact_configs,
      source_config,
  ) -> (
      get_messages_module().DeveloperconnectProjectsLocationsInsightsConfigsCreateRequest
  ):
    """Creates the insight config request for project scope."""
    projects = name.parse_target_projects(target_projects)
    dependent_projects = set(projects)
    # Get the artifact configs from CAIS for the target projects.
    cais_artifact_configs_dict = {}
    for project in projects:
      cais_artifacts = self.get_artifact_configs_from_cais_for_project(project)
      cais_artifact_configs_dict.update(cais_artifacts)

    user_artifact_configs_dict = name.parse_artifact_configs(
        user_artifact_configs
    )
    merged_artifact_configs_dict = self.merge_artifact_configs(
        cais_artifact_configs_dict, user_artifact_configs_dict
    )

    # Get the artifact configs and add new projects to the dependent projects
    # set.
    artifact_projects, artifact_configs = self.build_artifact_configs(
        merged_artifact_configs_dict, cais_artifact_configs_dict
    )
    # Add the artifact projects to the dependent projects set.
    dependent_projects.update(artifact_projects)
    # Get the P4SA and grant IAM roles to it.
    if dependent_projects:
      # Sort projects for deterministic behavior, important for tests.
      self.init_service_account(
          insight_config_ref.projectsId,
          sorted(dependent_projects),
          management_project=False,
      )
    if source_config:
      return self.messages.DeveloperconnectProjectsLocationsInsightsConfigsCreateRequest(
          parent=insight_config_ref.Parent().RelativeName(),
          insightsConfigId=insight_config_ref.insightsConfigsId,
          insightsConfig=self.messages.InsightsConfig(
              name=insight_config_ref.RelativeName(),
              projects=self.messages.Projects(projectIds=projects),
              artifactConfigs=artifact_configs,
              sourceConfigs=[self.messages.SourceConfig(
                  gitRepositoryLink=source_config.get('git-repository-link'),
              )],
          ),
      )
    else:
      return self.messages.DeveloperconnectProjectsLocationsInsightsConfigsCreateRequest(
          parent=insight_config_ref.Parent().RelativeName(),
          insightsConfigId=insight_config_ref.insightsConfigsId,
          insightsConfig=self.messages.InsightsConfig(
              name=insight_config_ref.RelativeName(),
              projects=self.messages.Projects(projectIds=projects),
              artifactConfigs=artifact_configs,
          ),
      )

  def create_source_config_insights_config_request(
      self,
      insight_config_ref,
      user_artifact_configs,
      source_config,
  ) -> (
      get_messages_module().DeveloperconnectProjectsLocationsInsightsConfigsCreateRequest
  ):
    """Creates the insight config request for source-only scope."""
    # Since there's no runtime context, CAIS discovery is skipped.
    cais_artifact_configs_dict = {}
    user_artifact_configs_dict = name.parse_artifact_configs(
        user_artifact_configs
    )

    merged_artifact_configs_dict = self.merge_artifact_configs(
        cais_artifact_configs_dict, user_artifact_configs_dict
    )
    artifact_projects, artifact_configs = self.build_artifact_configs(
        merged_artifact_configs_dict, cais_artifact_configs_dict
    )

    if artifact_projects:
      self.init_service_account(
          insight_config_ref.projectsId,
          sorted(artifact_projects),
          management_project=False,
      )

    insights_config_msg = self.messages.InsightsConfig(
        name=insight_config_ref.RelativeName(),
        artifactConfigs=artifact_configs,
    )

    # Populate source configs if the flag was provided.
    if source_config:
      insights_config_msg.sourceConfigs = [
          self.messages.SourceConfig(
              gitRepositoryLink=source_config.get('git-repository-link'),
          )
      ]

    return self.messages.DeveloperconnectProjectsLocationsInsightsConfigsCreateRequest(
        parent=insight_config_ref.Parent().RelativeName(),
        insightsConfigId=insight_config_ref.insightsConfigsId,
        insightsConfig=insights_config_msg,
    )

  def merge_artifact_configs(
      self,
      artifact_configs_dict: Dict[str, get_messages_module().ArtifactConfig],
      user_provided_artifact_configs: Dict[str, str],
  ) -> Dict[str, get_messages_module().ArtifactConfig]:
    """Merges artifact configs from CAIS and user provided configs user provided configs will overwrite configs extracted from CAIS if URIs match.
    """
    merged_artifact_configs_dict: Dict[
        str, get_messages_module().ArtifactConfig
    ] = {}
    # First, populate with CAIS-discovered configs
    if artifact_configs_dict:
      for uri, config_msg in artifact_configs_dict.items():
        merged_artifact_configs_dict[uri] = config_msg

    if not user_provided_artifact_configs:
      return merged_artifact_configs_dict

    for uri, build_project in user_provided_artifact_configs.items():
      # Create a new ArtifactConfig message to populate from user_config_data
      merged_artifact_configs_dict[uri] = self.messages.ArtifactConfig(
          uri=uri,
          googleArtifactAnalysis=self.messages.GoogleArtifactAnalysis(
              projectId=build_project
          ),
      )

    return merged_artifact_configs_dict

  def find_apphub_resources_and_grant_sa_permissions(
      self, insight_config_ref, app_hub_application
  ):
    """Finds the GKE workloads and Cloud Run services and grants SA permissions at the folder level for management project or returns the dependent projects for non-management projects.

    Args:
      insight_config_ref: The insight config reference.
      app_hub_application: The app hub application.

    Returns:
      A tuple of dependent projects(based on if it is a management project or
      not), gke workloads and cloud run services.
    """
    dependent_projects, gke_workloads, cloud_run_services = self.get_runtimes(
        app_hub_application.resource_name()
    )
    # If the app hub application is not a management project, return the
    # dependent projects from the runtime configs, we'll grant permissions to
    # this set of projects later.
    if not name.is_management_project(app_hub_application.project_id):
      return dependent_projects, gke_workloads, cloud_run_services

    # Management project, assign permissions to the folder and we don't need
    # dependent projects here.
    self.assign_management_permissions(insight_config_ref, app_hub_application)
    return set(), gke_workloads, cloud_run_services

  def get_artifact_configs_from_cais(
      self, gke_workloads: List[Any], cloud_run_services: List[Any]
  ) -> Dict[str, get_messages_module().ArtifactConfig]:
    """Queries CAIS for artifacts associated with the gke workloads in the resources scope.

    Args:
      gke_workloads: A list of GKE workloads.
      cloud_run_services: A list of Cloud Run services.

    Returns:
      A dict of artifact configs IC type.
    """
    # Use a dict to deduplicate artifact configs and allow users to overwrite
    # build projects.
    artifact_configs_dict: Dict[str, get_messages_module().ArtifactConfig] = {}
    for gke_workload in gke_workloads:
      assets = discover_artifacts.query_cais_for_gke_assets(gke_workload)
      artifact_uris = discover_artifacts.get_artifact_uris_from_gke_assets(
          assets
      )
      artifact_configs_dict.update(
          self.update_artifact_config_dict_from_artifact_uri(
              artifact_uris, artifact_configs_dict
          )
      )
    for cloud_run_service in cloud_run_services:
      assets = discover_artifacts.query_cais_for_cloud_run_services(
          cloud_run_service
      )
      artifact_uris = (
          discover_artifacts.get_artifact_uris_from_cloud_run_assets(assets)
      )
      artifact_configs_dict.update(
          self.update_artifact_config_dict_from_artifact_uri(
              artifact_uris, artifact_configs_dict
          )
      )
    return artifact_configs_dict

  def get_artifact_configs_from_cais_for_project(
      self, project: str
  ) -> Dict[str, get_messages_module().ArtifactConfig]:
    """Queries CAIS for artifacts associated with the GCP Project.

    Args:
      project: A GCP Project.

    Returns:
      A dict of artifact configs IC type.
    """
    artifact_configs_dict: Dict[str, get_messages_module().ArtifactConfig] = {}

    # Get GKE assets and extract artifact URIs.
    gke_assets = discover_artifacts.query_cais_for_gke_assets_in_project(
        project
    )
    gke_artifact_uris = discover_artifacts.get_artifact_uris_from_gke_assets(
        gke_assets
    )
    artifact_configs_dict.update(
        self.update_artifact_config_dict_from_artifact_uri(
            gke_artifact_uris, artifact_configs_dict
        )
    )

    # Get Cloud Run assets and extract artifact URIs.
    cloud_run_assets = (
        discover_artifacts.query_cais_for_cloud_run_services_in_project(project)
    )
    cloud_run_artifact_uris = (
        discover_artifacts.get_artifact_uris_from_cloud_run_assets(
            cloud_run_assets
        )
    )
    artifact_configs_dict.update(
        self.update_artifact_config_dict_from_artifact_uri(
            cloud_run_artifact_uris, artifact_configs_dict
        )
    )
    return artifact_configs_dict

  def update_artifact_config_dict_from_artifact_uri(
      self,
      artifact_uris: List[str],
      artifact_configs_dict: Dict[str, get_messages_module().ArtifactConfig],
  ) -> Dict[str, get_messages_module().ArtifactConfig]:
    """Updates the artifact configs dict with the artifact uris.

    Args:
      artifact_uris: The artifact uris.
      artifact_configs_dict: The artifact configs dict.

    Returns:
      The artifact configs dict.
    """
    for artifact in artifact_uris:
      validated_artifact_uri = name.validate_artifact_uri(artifact)
      # Skip if the artifact URI is invalid or the project is excluded.
      if not validated_artifact_uri:
        continue
      if validated_artifact_uri.project_id in _EXCLUDED_PROJECTS:
        continue
      base_uri = validated_artifact_uri.base_uri
      artifact_configs_dict[base_uri] = self.messages.ArtifactConfig(
          uri=base_uri,
          googleArtifactAnalysis=self.messages.GoogleArtifactAnalysis(
              projectId=validated_artifact_uri.project_id
          ),
      )
    return artifact_configs_dict

  def build_artifact_configs(
      self,
      merged_artifact_configs_dict: Dict[
          str, get_messages_module().ArtifactConfig
      ],
      cais_artifact_configs_dict: Dict[
          str, get_messages_module().ArtifactConfig
      ],
  ) -> Tuple[Set[str], List[get_messages_module().ArtifactConfig]]:
    """Builds the artifact configs and returns the dependent projects and artifact configs.

    Args:
      merged_artifact_configs_dict: A combined dict of artifact configs IC type
        from CAIS and user provided configs.
      cais_artifact_configs_dict: A dict of artifact configs IC type from CAIS.

    Returns:
      A tuple of dependent projects and artifact configs.
    """
    dependent_projects: Set[str] = set()
    # Print existing artifact configs if they exist.
    if not merged_artifact_configs_dict:
      log.status.Print('No existing artifact configurations found.')
      return dependent_projects, []
    # Prompt for build projects ONLY if we have CAIS-discovered
    # artifact configs.
    if cais_artifact_configs_dict:
      for artifact_config in merged_artifact_configs_dict.values():
        build_project = artifact_config.googleArtifactAnalysis.projectId
        log.status.Print(
            '\nBuild project'
            f' [{build_project}] will be'
            ' used to extract provenance information for artifact'
            f' [{artifact_config.uri}]'
        )
        change_build_project = (
            console_io.CanPrompt()
            and console_io.PromptContinue(
                prompt_string='Would you like to change the build project?',
                default=False,
            )
        )
        if change_build_project:
          build_project = self.prompt_for_build_project(artifact_config.uri)

        merged_artifact_configs_dict[artifact_config.uri] = (
            self.messages.ArtifactConfig(
                uri=artifact_config.uri,
                googleArtifactAnalysis=self.messages.GoogleArtifactAnalysis(
                    projectId=build_project
                ),
            )
        )

    # Return the dependent projects and the artifact configs.
    artifact_configs = list(merged_artifact_configs_dict.values())
    # Add the dependent build projects to the dependent projects list. We do
    # this here because users could overwrite build project selections while
    # prompting for artifact configs.
    dependent_projects.update(
        artifact.googleArtifactAnalysis.projectId
        for artifact in artifact_configs
    )
    return dependent_projects, artifact_configs

  def prompt_for_build_project(self, artifact_uri: str) -> str:
    """Prompts the user for the build project."""
    found = False
    build_project = None
    while not found:
      build_project = console_io.PromptResponse(
          'Please enter the build project for your artifact'
          f' [{artifact_uri}]: '
      )
      if not build_project:
        continue

      try:
        name.validate_project(build_project)
        found = True
      except apitools_exceptions.HttpForbiddenError:
        log.status.Print(
            'Permission denied when checking build project [{}]. Please'
            ' ensure your account has necessary permissions '
            'or that the project exists.'
            .format(build_project)
        )
      except apitools_exceptions.HttpBadRequestError:
        log.status.Print(
            'Invalid build project ID [{}]. Please ensure it is a valid'
            ' project ID (e.g., "my-project-123")'.format(build_project)
        )
      except exceptions.Error as e:
        log.warning(f'Error validating build project [{build_project}]: {e}')
    return build_project

  def get_gke_workloads_runtime(self, app_hub):
    """Gets the runtime configs.

    Args:
      app_hub: The app hub application.

    Returns:
      A tuple of gke runtime configs projects and gke workloads associated with
      the app hub application.
    """
    runtime_configs_projects = set()
    gke_workloads = []
    client = discover_apphub.DiscoveredApphubClient()
    workloads = client.list_workloads(
        page_size=100,
        parent=app_hub,
    )

    for workload in workloads:
      # Skip if these fields are not set.
      # This could happen if the workload is detached.
      if not workload.workloadReference or not workload.workloadReference.uri:
        continue
      gke_workload = name.parse_gke_deployment_uri(
          workload.workloadReference.uri
      )
      if not gke_workload:
        continue
      # Add the *project id* to the runtime configs projects set, ensuring it
      # is project id and not project number so that we don't add the same
      # project multiple times.
      runtime_configs_projects.add(
          name.Project(
              gke_workload.gke_namespace.gke_cluster.project
          ).project_id
      )
      gke_workloads.append(gke_workload)
    return runtime_configs_projects, gke_workloads

  def get_cloud_run_services_runtime(self, app_hub):
    """Gets the cloud run services.

    Args:
      app_hub: The app hub application.

    Returns:
      A tuple of cloud run runtime configs projects and cloud run services
      associated with the app hub application.
    """
    runtime_configs_projects = set()
    cloud_run_services = []
    client = discover_apphub.DiscoveredApphubClient()
    services = client.list_services(
        page_size=100,
        parent=app_hub,
    )
    for service in services:
      # Skip if these fields are not set.
      # This could happen if the service is detached.
      if not service.serviceReference or not service.serviceReference.uri:
        continue
      cloud_run_service = name.parse_cloud_run_service_uri(
          service.serviceReference.uri
      )
      if not cloud_run_service:
        continue
      # Add the *project id* to the runtime configs projects set, ensuring it
      # is project id and not project number so that we don't add the same
      # project multiple times.
      runtime_configs_projects.add(cloud_run_service.project_id)
      cloud_run_services.append(cloud_run_service)
    return runtime_configs_projects, cloud_run_services

  def get_runtimes(self, app_hub):
    """Gets the runtime configs.

    Args:
      app_hub: The app hub application.

    Returns:
      A tuple of runtime configs projects, gke workloads, and cloud run services
      associated with the app hub application.
    """
    if not app_hub:
      return set(), [], []
    gke_runtime_configs_projects, gke_workloads = (
        self.get_gke_workloads_runtime(app_hub)
    )
    cloud_run_runtime_configs_projects, cloud_run_services = (
        self.get_cloud_run_services_runtime(app_hub)
    )
    runtime_configs_projects = gke_runtime_configs_projects
    runtime_configs_projects.update(cloud_run_runtime_configs_projects)
    return runtime_configs_projects, gke_workloads, cloud_run_services

  def assign_management_permissions(self, insight_config_ref, app_hub):
    """Assigns permissions to at the folder level for management project."""
    # Management project, get permissions on the folder.
    folder_number = projects_api.Get(
        projects_util.ParseProject(app_hub.project_id)
    ).parent.id
    dependent_folder = [folder_number]
    self.init_service_account(
        insight_config_ref.projectsId,
        dependent_folder,
        management_project=True,
    )
    return

  def update(
      self,
      insight_config_ref: resources.Resource,
      *,
      discovery: bool,
      build_project: Optional[str],
      artifact_uri: Optional[str],
      app_hub: Optional[str],
      target_projects: Optional[List[str]],
  ) -> get_messages_module().Operation:
    """Updates the insight config."""
    old_insights_config = self.get_existing_insights_config(
        insight_config_ref
    )

    # Validate before granting IAM permissions.
    if app_hub and (
        old_insights_config.appHubApplication or old_insights_config.projects
    ):
      raise exceptions.Error(
          'Cannot update App Hub application because a runtime context'
          ' is already set for this Insight Config.'
      )
    if target_projects and (
        old_insights_config.appHubApplication or old_insights_config.projects
    ):
      raise exceptions.Error(
          'Cannot update target projects because a runtime context'
          ' is already set for this Insight Config.'
      )

    # Start with a copy of the old config to preserve existing fields.
    new_insights_config = self.messages.InsightsConfig(
        name=old_insights_config.name,
        appHubApplication=old_insights_config.appHubApplication,
        projects=old_insights_config.projects,
        artifactConfigs=old_insights_config.artifactConfigs,
        sourceConfigs=old_insights_config.sourceConfigs,
        runtimeConfigs=old_insights_config.runtimeConfigs,
        state=old_insights_config.state,
    )

    # Handle artifact config updates.
    if artifact_uri and build_project:
      self._update_artifact_config(
          new_insights_config, artifact_uri, build_project
      )

    # Handle context updates.
    if app_hub:
      app_hub_application = name.parse_app_hub_application_uri(app_hub)
      new_insights_config.appHubApplication = (
          app_hub_application.resource_name()
      )
      new_insights_config.projects = None
      discovery = True  # Any context addition requires discovery
    elif target_projects:
      projects = name.parse_target_projects(target_projects)
      new_insights_config.projects = self.messages.Projects(projectIds=projects)
      new_insights_config.appHubApplication = None
      discovery = True

    if discovery:
      new_insights_config.state = (
          self.messages.InsightsConfig.StateValueValuesEnum.PENDING
      )
    # IAM Permission Handling
    dependent_projects = set(self.get_dependent_projects(new_insights_config))

    effective_app_hub = new_insights_config.appHubApplication
    effective_app_hub_project = (
        name.extract_project(effective_app_hub) if effective_app_hub else None
    )

    if (
        effective_app_hub
        and effective_app_hub_project
        and name.is_management_project(effective_app_hub_project)
    ):
      folder_number = projects_api.Get(
          projects_util.ParseProject(effective_app_hub_project)
      ).parent.id
      self.init_service_account(
          insight_config_ref.projectsId,
          [folder_number],
          management_project=True,
      )
    elif dependent_projects:
      self.init_service_account(
          insight_config_ref.projectsId,
          sorted(dependent_projects),
          management_project=False,
      )

    update_request = self.messages.DeveloperconnectProjectsLocationsInsightsConfigsPatchRequest(
        insightsConfig=new_insights_config,
        name=insight_config_ref.RelativeName(),
    )
    return self.client.projects_locations_insightsConfigs.Patch(
        request=update_request
    )

  def _update_artifact_config(
      self, insights_config, artifact_uri, build_project
  ):
    """Updates or adds an artifact config to the insights_config message."""
    artifact_project = name.validate_artifact_uri(artifact_uri)
    if not artifact_project:
      raise exceptions.Error(
          f'Invalid artifact URI: {artifact_uri}. Artifact URI must be in the'
          ' format '
          '{location}-docker.pkg.dev/{project}/{repository}/{package}.'
      )
    # Check if the build project exists.
    try:
      name.validate_project(build_project)
    # Catch specific API errors first
    except apitools_exceptions.HttpForbiddenError:
      # Specific handling for permission errors
      raise exceptions.Error(
          'Permission denied when checking build project [{}]. Please ensure'
          ' the account [{}] has necessary permissions (e.g.,'
          ' resourcemanager.projects.get) or that the project exists.'
          .format(build_project, iam_util.GetAuthenticatedAccount())
      )
    except apitools_exceptions.HttpBadRequestError as e:
      raise exceptions.Error(
          'Invalid build project ID [{}]: {}. Please ensure it is a valid'
          ' project ID (e.g., "my-project-123") and not an artifact URI.'
          .format(build_project, e)
      )

    updated_artifact = self.messages.ArtifactConfig(
        uri=artifact_uri,
        googleArtifactAnalysis=self.messages.GoogleArtifactAnalysis(
            projectId=build_project
        ),
    )

    found = False
    if insights_config.artifactConfigs:
      for index, artifact_config in enumerate(insights_config.artifactConfigs):
        if artifact_config.uri == artifact_uri:
          insights_config.artifactConfigs[index] = updated_artifact
          found = True
          break
    if not found:
      # Add a new artifact config since it doesn't exist.
      if insights_config.artifactConfigs is None:
        insights_config.artifactConfigs = []
      insights_config.artifactConfigs.append(updated_artifact)

  def insights_config_message_type(self, current_insights_config):
    """Creates a new insights config message type."""
    return self.messages.InsightsConfig(
        state=current_insights_config.state,
        artifactConfigs=current_insights_config.artifactConfigs,
        runtimeConfigs=current_insights_config.runtimeConfigs,
        sourceConfigs=current_insights_config.sourceConfigs,
    )

  def get_existing_insights_config(self, insight_config_ref):
    """Gets the insight config."""
    try:
      return self.client.projects_locations_insightsConfigs.Get(
          request=self.messages.DeveloperconnectProjectsLocationsInsightsConfigsGetRequest(
              name=insight_config_ref.RelativeName(),
          )
      )
    except apitools_exceptions.HttpNotFoundError:
      raise exceptions.Error(
          f'Insights Config [{insight_config_ref.insightsConfigsId}] not found'
          f' in project [{insight_config_ref.projectsId}] location'
          f' [{insight_config_ref.locationsId}].'
      )

  def get_dependent_projects(
      self, insights_config: get_messages_module().InsightsConfig
  ) -> List[str]:
    """Gets the P4SA projects for the insight config."""
    projects = set()
    app_hub_project = (
        name.extract_project(insights_config.appHubApplication)
        if insights_config.appHubApplication
        else None
    )
    if app_hub_project:
      projects.add(app_hub_project)
    elif insights_config.projects and insights_config.projects.projectIds:
      projects.update(insights_config.projects.projectIds)

    for artifact_config in insights_config.artifactConfigs:
      if artifact_config.uri:
        artifact_uri = name.validate_artifact_uri(artifact_config.uri)
        if artifact_uri:
          projects.add(artifact_uri.project_id)
      if (
          artifact_config.googleArtifactAnalysis
          and artifact_config.googleArtifactAnalysis.projectId
      ):
        projects.add(artifact_config.googleArtifactAnalysis.projectId)

    for runtime_config in insights_config.runtimeConfigs:
      if runtime_config.uri:
        projects.add(name.extract_project(runtime_config.uri))
    # Sort projects for deterministic behavior.
    return sorted(projects)

  def init_service_account(
      self, p4sa_project, dependent_resources, management_project
  ):
    """Get the Developer Connect P4SA, and grant IAM roles to it.

    1) First, get the P4SA for the project.
    Args:
      p4sa_project: The project to get the P4SA for.
      dependent_resources: The resources to grant the P4SA permissions to.
      management_project: Whether the resource is a management project.

    2) Then grant necessary roles needed to the P4SA for updating an insight
      config.

    3) If the app hub application is a management project, grant the P4SA
      permissions on the folder.

    4) If the app hub application is a non management project, grant the P4SA
      permissions on the dependent projects.

    Raises:
      InsightsConfigInitializationError: P4SA failed to be created.
    """
    service_name = common.GetApiServiceName(self.api_version)
    if self.p4sa_email is None:
      self.p4sa_email = _get_p4sa(p4sa_project, service_name)
    if not self.p4sa_email:
      raise InsightsConfigInitializationError(
          'Failed to get P4SA for project {}.'.format(p4sa_project)
      )
    if management_project:
      if len(dependent_resources) == 1:
        self.bind_roles_to_service_account(
            self.p4sa_email, dependent_resources[0], True
        )
      else:
        log.warning(
            'Could not find folder number for the management project.'
            'Skipping permissions checks/binding.'
        )
    else:
      # Sort the projects to ensure deterministic behavior for tests.
      for project in sorted(dependent_resources):
        project_ref = projects_util.ParseProject(project)
        self.bind_roles_to_service_account(self.p4sa_email, project_ref, False)

  def bind_roles_to_service_account(
      self, sa_email, resource_ref, management_project
  ):
    """Binds roles to the provided service account.

    Args:
      sa_email: str, the service account to bind roles to.
      resource_ref: str, the resource to bind roles to.
      management_project: bool, whether the resource is a management project.
    """
    for role in _ROLES:
      self.prompt_to_bind_role_if_missing(
          sa_email,
          resource_ref,
          role,
          management_project,
          reason='required to update the Developer Connect Insights Config',
      )

  def prompt_to_bind_role_if_missing(
      self, sa_email, resource_ref, role, management_project, reason=''
  ):
    """Prompts to bind the role to the service account in project level if missing.

    If the console cannot prompt, a warning is logged instead.

    Args:
      sa_email: The service account email to bind the role to.
      resource_ref: The resource to bind the role to.
      role: The role to bind if missing.
      management_project: Whether the resource is a management project.
      reason: Extra information to print explaining why the binding is
        necessary.
    """
    member = 'serviceAccount:{}'.format(sa_email)
    try:
      if management_project:
        iam_policy = folders.GetIamPolicy(resource_ref)
      else:
        iam_policy = projects_api.GetIamPolicy(resource_ref)
      if self.has_role_binding(iam_policy, sa_email, role):
        return

      log.status.Print(
          '\nService account [{}] is missing the role [{}] in project [{}].\n{}'
          .format(sa_email, role, resource_ref, reason)
      )

      bind = console_io.CanPrompt() and console_io.PromptContinue(
          prompt_string=(
              '\nBind the role [{}] to service account [{}] in project [{}]?'
              .format(role, sa_email, resource_ref)
          )
      )
      if not bind:
        log.warning('Manual binding of above role will be necessary.\n')
        return

      if management_project:
        messages = folders.FoldersMessages()
        iam_util.AddBindingToIamPolicy(
            messages.Binding, iam_policy, member, role
        )
        folders.SetIamPolicy(resource_ref, iam_policy)
      else:
        projects_api.AddIamPolicyBinding(resource_ref, member, role)

      log.status.Print(
          'Successfully bound the role [{}] to service account [{}] in project'
          ' [{}]'.format(role, sa_email, resource_ref)
      )
    except apitools_exceptions.HttpForbiddenError:
      log.warning(
          (
              'Your account does not have permission to check or bind IAM'
              ' policies to resource [%s]. If the deployment fails, ensure [%s]'
              ' has the role [%s] before retrying.'
          ),
          resource_ref,
          sa_email,
          role,
      )

  def has_role_binding(self, iam_policy, sa_email, role):
    """Returns whether the given SA has the given role bound in given policy.

    Args:
      iam_policy: The IAM policy to check.
      sa_email: The service account to check.
      role: The role to check for.
    """
    return any(
        'serviceAccount:{}'.format(sa_email) in b.members and b.role == role
        for b in iam_policy.bindings
    )

  def get_operation_ref(self, operation):
    """Converts an operation to a resource that can be used with `waiter.WaitFor`."""
    return self._resource_parser.ParseRelativeName(
        operation.name, 'developerconnect.projects.locations.operations')

  def wait_for_operation(
      self,
      operation_ref,
      message,
      has_result=True,
      max_wait=datetime.timedelta(seconds=600),
  ):
    """Waits for a Developer Connect operation to complete.

      Polls the Developer Connect Insights Operation service until the operation
      completes, fails, or max_wait_seconds elapses.

    Args:
      operation_ref: a resource reference created by GetOperationRef describing
        the operation.
      message: a message to display to the user while they wait.
      has_result: If True, the function will return the target of the operation
        (i.e. the InsightConfig) when it completes. If False, nothing will be
        returned (useful for Delete operations).
      max_wait: The time to wait for the operation to complete before returning.

    Returns:
      A resource reference to the target of the operation if has_result is True,
      otherwise None.
    """
    if has_result:
      poller = waiter.CloudOperationPoller(
          self.client.projects_locations_insightsConfigs,
          self.client.projects_locations_operations,
      )
    else:
      poller = waiter.CloudOperationPollerNoResources(
          self.client.projects_locations_operations
      )

    return waiter.WaitFor(
        poller, operation_ref, message, max_wait_ms=max_wait.seconds * 1000
    )



# -*- coding: utf-8 -*- #
# Copyright 2026 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Command to generate optimized Kubernetes manifests for a workload profile."""

from googlecloudsdk.api_lib.ai.recommender import util
from googlecloudsdk.calliope import arg_parsers
from googlecloudsdk.calliope import base
from googlecloudsdk.core import exceptions
from googlecloudsdk.core import log
from googlecloudsdk.core.resource import resource_printer
from googlecloudsdk.core.resource import resource_printer_base
from googlecloudsdk.core.resource import resource_projector
from googlecloudsdk.core.util import files


@base.DefaultUniverseOnly
@base.ReleaseTracks(base.ReleaseTrack.GA)
class Create(base.Command):
  """Generate optimized Kubernetes manifests for a given workload profile."""

  @staticmethod
  def Args(parser):
    """Register flags for this command."""
    parser.add_argument(
        "--workload",
        type=str,
        required=True,
        help="""\
        The name of the optimization set to generate the manifest for.
        This specifies the workload, workload version, and workload
        characterization to optimize for (e.g., "redis-7-caching").
        """,
    )
    parser.add_argument(
        "--cluster-version",
        type=str,
        required=True,
        help="The GKE version to generate the manifest for.",
    )
    parser.add_argument(
        "--options",
        metavar="KEY=VALUE",
        type=arg_parsers.ArgDict(),
        help=(
            "Additional key-value pair options for generating the manifest. "
            "For example, to specify allowed machine types: "
            "--options=machineType=type1,type2"
        ),
        required=False,
    )
    parser.add_argument(
        "--output",
        choices=["manifest", "all"],
        default="all",
        help="The output to display. Default is all.",
    )
    parser.add_argument(
        "--output-path",
        help=(
            "The path to save the output to. If not specified, output to the"
            " terminal."
        ),
    )
    parser.display_info.AddFormat("manifests")
    resource_printer.RegisterFormatter("manifests", ManifestPrinter)

  def GetOptionClass(self, messages):
    """Returns the track-specific Option class."""
    return messages.Option

  def GetRequestClass(self, messages):
    """Returns the track-specific request class."""
    return messages.GenerateOptimizationSetManifestRequest

  def GetClientVersion(self, client):
    """Returns the track-specific client version."""
    return client.v1

  def Run(self, args):
    """This is what gets called when the user runs this command."""
    track = self.ReleaseTrack()
    client = util.GetClientInstance(track)
    messages = util.GetMessagesModule(track)

    option_class = self.GetOptionClass(messages)
    options = (
        [option_class(key=k, value=v) for k, v in args.options.items()]
        if args.options
        else []
    )

    request_class = self.GetRequestClass(messages)
    req = request_class(
        optimizationSet=args.workload,
        clusterVersion=args.cluster_version,
        options=options,
    )

    version_client = self.GetClientVersion(client)
    resp = version_client.GenerateOptimizationSetManifest(req)

    if args.output_path:
      _SaveOutputToFile(resp, args, args.output_path)

    return resp


def _SaveOutputToFile(resp, args, output_path):
  """Saves the output content to a file."""
  try:
    with files.FileWriter(output_path) as f:
      if args.output in ("manifest", "all"):
        resource_printer.Print(
            resource_projector.MakeSerializable(resp),
            "manifests",
            out=f,
            single=True,
        )
    log.status.Print(f"Output saved to {output_path}")
  except exceptions.Error as e:
    log.error(f"An error occurred while saving output to file: {e}")


class ManifestPrinter(resource_printer_base.ResourcePrinter):
  """Renders Kubernetes manifests."""

  def _AddRecord(self, record, delimit=True):
    """Renders the resource output."""
    if not record:
      return

    parts = []
    for manifest in record.get("kubernetesManifests", []):
      parts.extend([manifest.get("content", ""), "---"])

    if parts:
      self._out.write("\n".join(parts) + "\n")


@base.DefaultUniverseOnly
@base.ReleaseTracks(base.ReleaseTrack.ALPHA)
@base.Hidden
class CreateAlpha(Create):
  """Generate optimized Kubernetes manifests for a given workload profile."""

  def GetOptionClass(self, messages):
    """Returns the track-specific Option class."""
    return messages.GoogleCloudGkerecommenderV1alpha1Option

  def GetRequestClass(self, messages):
    """Returns the track-specific request class."""
    return (
        messages.GoogleCloudGkerecommenderV1alpha1GenerateOptimizationSetManifestRequest
    )

  def GetClientVersion(self, client):
    """Returns the track-specific client version."""
    return client.v1alpha1
